import pandas as pd
class WebullStockData:
    """A class representing stock data obtained from Webull.

    Attributes:
        web_name (str): The name of the stock.
        web_symb (str): The stock's symbol.
        web_exchange (str): The exchange code where the stock is traded.
        web_stock_close (float): The stock's closing price.
        last_earnings (str): The date of the stock's latest earnings report.
        web_stock_vol (int): The stock's trading volume.
        web_change_ratio (float): The stock's price change ratio.
        web_stock_open (float): The stock's opening price.
        web_stock_high (float): The stock's highest price.
        web_stock_low (float): The stock's lowest price.
        fifty_high (float): The stock's 52-week high price.
        avg_vol3m (float): The stock's average trading volume over the past 3 months.
        fifty_low (float): The stock's 52-week low price.
        avg_10d_vol (float): The stock's average trading volume over the past 10 days.
        outstanding_shares (int): The number of outstanding shares of the stock.
        total_shares (int): The total number of shares of the stock.
        estimated_earnings (str): The estimated date of the stock's next earnings report.
        web_vibrate_ratio (float): The stock's price fluctuation ratio.
    """
    def __init__(self, r):
        self.r = r
        for i in self.r:
            self.web_name = i.get("name", None)
            self.web_symb = i.get("symbol", None)
            self.web_exchange = i.get("disExchangeCode", None)
            self.web_stock_close =i.get("close", None)
            self.last_earnings = i.get('latestEarningsDate',None)
            self.web_stock_vol =i.get("volume",None)
            self.web_change_ratio = i.get("changeRatio", None)
            self.web_stock_open =i.get("open",None)
            self.web_stock_high =i.get("high", None)
            self.web_stock_low =i.get("low", None)
            self.fifty_high = i.get("fiftyTwoWkHigh", None)
            self.avg_vol3m = i.get('avgVol3M')
            self.fifty_low = i.get("fiftyTwoWkLow", None)
            self.avg_10d_vol = i.get("avgVol10D", None)
            self.outstanding_shares = i.get('outstandingShares', None)
            self.total_shares = i.get('totalShares', None)

            try:
                self.estimated_earnings = i.get("nextEarningDay", None)
                self.web_vibrate_ratio = i.get('vibrateRatio', None)
            except KeyError:
                self.estimated_earnings = None
                self.web_vibrate_ratio = None


            #self.data_dict = {
            #     'Company Name': i.get("name", None),
            #     'Symbol': i.get("symbol", None),
            #     'Exchange': i.get("disExchangeCode", None),
            #     'Close Price': i.get("close", None),
            #     'Latest Earnings': i.get('latestEarningsDate', None),
            #     'Volume': i.get("volume", None),
            #     'Change Ratio': i.get("changeRatio", None),
            #     'Open Price': i.get("open", None),
            #     'High Price': i.get("high", None),
            #     'Low Price': i.get("low", None),
            #     '52week High': i.get("fiftyTwoWkHigh", None),
            #     'Avg 3month Volume': i.get('avgVol3M', None),
            #     '52week Low': i.get("fiftyTwoWkLow", None),
            #     'Avg 10day Volume': self.r.get("avgVol10D", None),
            #     'Outstanding Shares': self.r.get('outstandingShares', None),
            #     'Total Shares': self.r.get('totalShares', None)
            # }

            # self.df = pd.DataFrame(self.data_dict, index=[0]).transpose()




        


import sys
from pathlib import Path

# Add the project directory to the sys.path
project_dir = str(Path(__file__).resolve().parents[1])
if project_dir not in sys.path:
    sys.path.append(project_dir)

import os

from dotenv import load_dotenv
load_dotenv()

from fudstop.apis.polygonio.mapping import option_condition_dict, OPTIONS_EXCHANGES
from list_sets.ticker_lists import most_active_tickers

from apis.helpers import flatten_dict
import aiohttp
import asyncio
from aiohttp import ClientTimeout
from asyncio import Semaphore
sema = Semaphore()

from datetime import datetime

YOUR_API_KEY = os.environ.get('YOUR_POLYGON_KEY')

async def process_universal_snapshot(data):
    batch = []
    for i in data:
        trade_timestamp = i.get('last_trade.sip_timestamp')
        if trade_timestamp is not None:
            trade_timestamp = datetime.fromtimestamp(trade_timestamp / 1e9)

        trade_conditions = i.get('last_trade.conditions', None)
        if trade_conditions is not None:
            trade_conditions = option_condition_dict.get(trade_conditions[0])

        data_dict = {
            'name': i.get('name'),
            'option_symbol': i.get('ticker'),
            'underlying_symbol': i.get('underlying_asset.ticker'),
            'strike': i.get('details.strike_price'),
            'call_put': i.get('details.contract_type'),
            'expiry': (i.get('details.expiration_date')),
            'underlying_price': i.get('underlying_asset.price'),
            'change': i.get('session.change'),
            'change_percent': i.get('session.change_percent'),
            'early_change': i.get('session.early_trading_change'),
            'early_change_percent': i.get('session.early_trading_change_percent'),
            'change_to_break_even': i.get('underlying_asset.change_to_break_even'),
            'break_even_price': i.get('break_even_price'),
            'open': i.get('session.open'),
            'high': i.get('session.high'),
            'low': i.get('session.low'),
            'close': i.get('session.close'),
            'previous_close': i.get('session.previous_close'),
            'volume': i.get('session.volume'),
            'oi': i.get('open_interest'),
            'iv': i.get('implied_volatility'),
            'delta': i.get('greeks.delta'),
            'gamma': i.get('greeks.gamma'),
            'theta': i.get('greeks.theta'),
            'vega': i.get('greeks.vega'),
            'trade_size': i.get('last_trade.size'),
            'trade_price': i.get('last_trade.price'),
            'trade_exchange': OPTIONS_EXCHANGES.get(i.get('last_trade.exchange')),
            'trade_conditions': trade_conditions,
            'trade_timestamp': trade_timestamp,
            'ask': i.get('last_quote.ask'),
            'ask_size': i.get('last_quote.ask_size'),
            'ask_exchange': OPTIONS_EXCHANGES.get(i.get('last_quote.ask_exchange')),
            'bid': i.get('last_quote.bid'),
            'bid_size': i.get('last_quote.bid_size'),
            'bid_exchange': OPTIONS_EXCHANGES.get(i.get('last_quote.bid_exchange'))}
                            



        if data_dict['expiry'] is not None:
            current_date = datetime.now()

            def parse_date(date_str):
                year, month, day = map(int, date_str.split('-'))
                return datetime(year, month, day)
            expiry_date = parse_date(data_dict['expiry'])
        if data_dict['underlying_price'] is not None and data_dict['strike'] is not None and data_dict['strike'] != 0:
            data_dict['moneyness'] = data_dict['underlying_price'] / data_dict['strike']
        else:
            data_dict['moneyness'] = None

        # Calculate Time to Expiry (in days)
        if data_dict['expiry'] is not None:
            try:
                expiry_date = datetime.strptime(data_dict['expiry'], '%Y-%m-%d')
                current_date = datetime.now()
                if expiry_date is not None and current_date is not None:
                    data_dict['time_to_expiry'] = (expiry_date - current_date).days
                else:
                    data_dict['time_to_expiry'] = None
            except ValueError:
                # Handle date parsing errors if needed
                data_dict['time_to_expiry'] = None
        else:
            data_dict['time_to_expiry'] = None
        # Calculate Intrinsic Value

        # Calculate Intrinsic Value
        if data_dict.get('underlying_price') is not None and data_dict.get('strike') is not None:
            data_dict['intrinsic_value_call'] = max(0, data_dict['underlying_price'] - data_dict['strike'])
            data_dict['intrinsic_value_put'] = max(0, data_dict['strike'] - data_dict['underlying_price'])
        else:
            data_dict['intrinsic_value_call'] = None
            data_dict['intrinsic_value_put'] = None

        # Calculate Extrinsic Value
        if all(k in data_dict and data_dict[k] is not None for k in ['close', 'intrinsic_value_call', 'intrinsic_value_put']):
            data_dict['extrinsic_value'] = data_dict['close'] - max(data_dict['intrinsic_value_call'], data_dict['intrinsic_value_put'])
        else:
            data_dict['extrinsic_value'] = None



        # Calculate Liquidity Score
        if data_dict.get('volume') is not None and data_dict.get('oi') is not None:
            data_dict['liquidity_score'] = data_dict['volume'] * data_dict['oi']
        else:
            data_dict['liquidity_score'] = None

        # Calculate Implied Leverage
        if data_dict.get('underlying_price') is not None and data_dict.get('close') is not None and data_dict['close'] != 0:
            data_dict['implied_leverage'] = data_dict['underlying_price'] / data_dict['close']
        else:
            data_dict['implied_leverage'] = None

        # Calculate Delta to Theta Ratio
        if data_dict.get('delta') is not None and data_dict.get('theta') is not None and data_dict['theta'] != 0:
            data_dict['delta_to_theta_ratio'] = data_dict['delta'] / data_dict['theta']
        else:
            data_dict['delta_to_theta_ratio'] = None

        # Calculate Cost of Theta using Close Price
        if data_dict.get('theta') is not None and data_dict.get('close') is not None and data_dict['close'] != 0:
            data_dict['cost_of_theta'] = data_dict['theta'] / data_dict['close']
        else:
            data_dict['cost_of_theta'] = None

        # # Calculate Risk-Reward Ratio
        # data_dict['risk_reward_ratio'] = data_dict['potential_profit'] / data_dict['potential_loss']


        batch.append(data_dict)

    if len(batch) == 250:
        return batch





async def get_universal_snapshot(ticker, retries=1): #✅
    """Fetches the Polygon.io universal snapshot API endpoint"""
    timeout = ClientTimeout(total=10)  # 10 seconds timeout for the request
    
    for retry in range(retries):
       # async with sema:
        url = f"https://api.polygon.io/v3/snapshot?ticker.any_of={ticker}&apiKey={YOUR_API_KEY}&limit=250"
        print(url)
        
        async with aiohttp.ClientSession(timeout=timeout) as session:
            try:
                async with session.get(url) as resp:
                    data = await resp.json()
                    results = data.get('results', None)
    
                    if results is not None:
                        flattened_results = [flatten_dict(result) for result in results]
                        return flattened_results
                        
            except aiohttp.ClientConnectorError:
                print("ClientConnectorError occurred. Retrying...")
                continue
            
            except aiohttp.ContentTypeError as e:
                print(f"ContentTypeError occurred: {e}")  # Consider logging this
                continue
            
            except Exception as e:
                print(f"An unexpected error occurred: {e}")  # Consider logging this
                continue



async def process_contract(data):
        
    data = await get_universal_snapshot(data)

    if data is not None:
        for contract in data:
            try:

                option_symbol = contract.get('ticker')
                expiry = contract.get('details').get('expiration_date')
                strike = contract.get('details').get('strike_price')
                call_put = contract.get('details').get('contract_type')
                underlying_symbol = contract.get('underlying_asset').get('ticker')
                underlying_price = contract.get('underlying_asset').get('price')
                change_to_breakeven = contract.get('underlying_asset').get('change_to_breakeven')

                

                trade_size = contract.get('last_trade').get('size')
                trade_price = contract.get('last_trade').get('price')
                trade_conditions = contract.get('last_trade', None).get('conditions', None)
                sip_timestamp = contract.get('last_trade', None).get('sip_timestamp', None)
                if sip_timestamp is not None:
                    sip_timestamp = datetime.fromtimestamp(sip_timestamp / 1e9)
                trade_conditions = [option_condition_dict.get(c) for c in trade_conditions] if trade_conditions is not None else []
                trade_conditions = trade_conditions[0]
                trade_exchange = OPTIONS_EXCHANGES.get(contract.get('last_trade').get('exchange', None))
                break_even_price = contract.get('break_even_price', None)
                implied_volatility = contract.get('implied_volatility', None)
                open_interest = contract.get('open_interest', None)
                name = contract.get('name', None)
                open = contract.get('session').get('open', None)
                high = contract.get('session').get('high', None)
                low = contract.get('session').get('low', None)
                close = contract.get('session').get('close', None)
                volume = contract.get('session').get('volume', None)
                prev_close = contract.get('session').get('previous_close')
                early_trading_change_percent = contract.get('session').get('early_trading_change_percent')
                change = contract.get('session').get('change')
                early_trading_change = contract.get('session').get('early_trading_change')
                change_percent = contract.get('session').get('change_percent', None)

                delta = contract.get('greeks').get('delta')
                vega = contract.get('greeks').get('vega')
                theta = contract.get('greeks').get('theta')
                gamma = contract.get('greeks').get('gamma')

                ask = contract.get('last_quote').get('ask')
                bid = contract.get('last_quote').get('bid')
                ask_size = contract.get('last_quote').get('ask_size')
                bid_size = contract.get('last_quote').get('bid_size')
                ask_exchange = OPTIONS_EXCHANGES.get(contract.get('last_quote').get('ask_exchange'))
                bid_exchange = OPTIONS_EXCHANGES.get(contract.get('last_quote').get('bid_exchange'))


                data_dict = {
                    'option_symbol': option_symbol,
                    'expiry': expiry,
                    'strike': strike,
                    'call_put': call_put,
                    'underlying_symbol': underlying_symbol,
                    'underlying_price': underlying_price,
                    'change_to_breakeven': change_to_breakeven,
                    'trade_size': trade_size,
                    'trade_price': trade_price,
                    'trade_conditions': trade_conditions,
                    'sip_timestamp': sip_timestamp,
                    'trade_exchange': trade_exchange,
                    'break_even_price': break_even_price,
                    'implied_volatility': implied_volatility,
                    'open_interest': open_interest,
                    'name': name,
                    'open': open,
                    'high': high,
                    'low': low,
                    'close': close,
                    'volume': volume,
                    'prev_close': prev_close,
                    'early_trading_change_percent': early_trading_change_percent,
                    'change': change,
                    'early_trading_change': early_trading_change,
                    'change_percent': change_percent,
                    'delta': delta,
                    'vega': vega,
                    'theta': theta,
                    'gamma': gamma,
                    'ask': ask,
                    'bid': bid,
                    'ask_size': ask_size,
                    'bid_size': bid_size,
                    'ask_exchange': ask_exchange,
                    'bid_exchange': bid_exchange
                }

            except Exception as e:
                # Handle any other exceptions that might occur
                print(f"An error occurred: {e}")

async def process_symbol(symbol):
    data = await get_universal_snapshot(symbol)

    if data is not None:
        return await asyncio.gather(*[process_contract(contract) for contract in data])
# async def main():
   
#     print("Database connected.")

#     tasks = []
#     batch_size = 8  # You can adjust this number based on your needs

#     for symbol in most_active_tickers:
#         print(f"Preparing task for symbol: {symbol}")
#         tasks.append(process_symbol(symbol))

#         if len(tasks) >= batch_size:
#             try:
#                 print(f"Executing batch of {batch_size} tasks.")
#                 await asyncio.gather(*tasks)
#                 print(f"Completed batch of {batch_size} tasks.")
#             except Exception as e:
#                 print(f"An error occurred while processing a batch: {e}")
#             finally:
#                 tasks.clear()  # Clear the list for the next batch, regardless of success or failure

#     # Handle any remaining tasks
#     if tasks:
#         try:
#             print(f"Executing final batch of {len(tasks)} tasks.")
#             await asyncio.gather(*tasks)
#             print("Completed final batch.")
#         except Exception as e:
#             print(f"An error occurred while processing the final batch: {e}")

#     print("All tasks completed.")

# asyncio.run(main())


from dataclasses import dataclass, asdict
from datetime import datetime

import pandas as pd

@dataclass
class HighestShortedData:
    rank: int
    ticker: str
    date: datetime
    short_interest: int
    average_volume: int
    days_to_cover: float
    percent_float_short: float


    def as_dataframe(self) -> pd.DataFrame:
        """Converts the object to a pandas DataFrame."""
        return pd.DataFrame([asdict(self)])
    

    def insert_into_pg(self, conn):
        cursor = conn.cursor()
        insert_query = """INSERT INTO highest_shorted (rank, ticker, date, short_interest, average_volume, days_to_cover, percent_float_short) VALUES (%s, %s, %s, %s, %s, %s, %s);"""
        cursor.execute(insert_query, (self.rank, self.ticker, self.date, self.short_interest, self.average_volume, self.days_to_cover, self.percent_float_short))
        conn.commit()



import os
import asyncpg
from asyncio import Lock
import pandas as pd
from datetime import datetime
import numpy as np
from dateutil.parser import parse

lock = Lock()

class DatabaseManager:
    def __init__(self, connection_string):
        self.conn = None
        self.pool = None
        self.connection_string = connection_string
        
        self.chat_memory = []  # In-memory list to store chat messages

    async def connect(self):
        self.pool = await asyncpg.create_pool(
            dsn=self.connection_string, min_size=1, max_size=10
        )

        return self.pool
    
    async def table_exists(self, table_name):
        query = f"SELECT EXISTS (SELECT FROM information_schema.tables WHERE table_name = '{table_name}');"
        async with self.pool.acquire() as conn:
            async with conn.transaction():
                exists = await conn.fetchval(query)
        return exists
    

    async def create_table(self, df, table_name):
        print("Connected to the database.")
        dtype_mapping = {
            'int64': 'BIGINT',
            'float64': 'DOUBLE PRECISION',
            'object': 'TEXT',
            'bool': 'BOOLEAN',
            'datetime.date': 'TIMESTAMP',
            'datetime.datetime': 'TIMESTAMP',
            'datetime64[ns]': 'timestamp',
            'datetime64[ms]': 'timestamp',
            'datetime64[ns, US/Eastern]': 'TIMESTAMP WITH TIME ZONE',
            'string': 'TEXT'
        }

        # Update dtype_mapping based on the data in the DataFrame
        dtype_mapping = await self.update_dtype_mapping(df, dtype_mapping)

        print(f"Updated DataFrame dtypes: {dtype_mapping}")

        # Check for large integers and update dtype_mapping accordingly
        for col, dtype in zip(df.columns, df.dtypes):
            if dtype == 'int64':
                max_val = df[col].max()
                min_val = df[col].min()
                if max_val > 2**31 - 1 or min_val < -2**31:
                    dtype_mapping['int64'] = 'BIGINT'
        history_table_name = f"{table_name}_history"
        async with self.pool.acquire() as connection:

            table_exists = await connection.fetchval(f"SELECT to_regclass('{table_name}')")
            
            if table_exists is None:
                create_query = f"""
                CREATE TABLE {table_name} (
                    {', '.join(f'"{col}" {dtype_mapping[str(dtype)]}' for col, dtype in zip(df.columns, df.dtypes))}
                )
                """

                print(f"Creating table with query: {create_query}")

                # Create the history table
                history_create_query = f"""
                CREATE TABLE IF NOT EXISTS {history_table_name} (
                    id serial PRIMARY KEY,
                    operation CHAR(1) NOT NULL,
                    changed_at TIMESTAMP NOT NULL DEFAULT current_timestamp,
                    {', '.join(f'"{col}" {dtype_mapping[str(dtype)]}' for col, dtype in zip(df.columns, df.dtypes))}
                );
                """
                print(f"Creating history table with query: {history_create_query}")
                await connection.execute(history_create_query)
                try:
                    await connection.execute(create_query)
                    print(f"Table {table_name} created successfully.")
                except asyncpg.UniqueViolationError as e:
                    print(f"Unique violation error: {e}")
            else:
                print(f"Table {table_name} already exists.")
            
            # Create the trigger function
            trigger_function_query = f"""
            CREATE OR REPLACE FUNCTION save_to_{history_table_name}()
            RETURNS TRIGGER AS $$
            BEGIN
                INSERT INTO {history_table_name} (operation, changed_at, {', '.join(f'"{col}"' for col in df.columns)})
                VALUES (
                    CASE
                        WHEN (TG_OP = 'DELETE') THEN 'D'
                        WHEN (TG_OP = 'UPDATE') THEN 'U'
                        ELSE 'I'
                    END,
                    current_timestamp,
                    {', '.join('OLD.' + f'"{col}"' for col in df.columns)}
                );
                RETURN NEW;
            END;
            $$ LANGUAGE plpgsql;
            """
            await connection.execute(trigger_function_query)

            # Create the trigger
            trigger_query = f"""
            DROP TRIGGER IF EXISTS tr_{history_table_name} ON {table_name};
            CREATE TRIGGER tr_{history_table_name}
            AFTER UPDATE OR DELETE ON {table_name}
            FOR EACH ROW EXECUTE FUNCTION save_to_{history_table_name}();
            """
            await connection.execute(trigger_query)


            # Alter existing table to add any missing columns
            for col, dtype in zip(df.columns, df.dtypes):
                alter_query = f"""
                DO $$
                BEGIN
                    BEGIN
                        ALTER TABLE {table_name} ADD COLUMN "{col}" {dtype_mapping[str(dtype)]};
                    EXCEPTION
                        WHEN duplicate_column THEN
                        NULL;
                    END;
                END $$;
                """
                await connection.execute(alter_query)


    async def batch_insert_dataframe(self, df, table_name, unique_columns, batch_size=250):
        async with lock:
            if not await self.table_exists(table_name):
                await self.create_table(df, table_name)
                
            df = df.copy()
            df.dropna(inplace=True)
            
            records = df.to_records(index=False)
            data = list(records)

            async with self.pool.acquire() as connection:
                column_types = await connection.fetch(
                    f"SELECT column_name, data_type FROM information_schema.columns WHERE table_name = '{table_name}'"
                )
                type_mapping = {col: next((item['data_type'] for item in column_types if item['column_name'] == col), None) for col in df.columns}

                async with connection.transaction():
                    insert_query = f"""
                    INSERT INTO {table_name} ({', '.join(f'"{col}"' for col in df.columns)}) 
                    VALUES ({', '.join('$' + str(i) for i in range(1, len(df.columns) + 1))})
                    ON CONFLICT ({unique_columns})
                    DO UPDATE SET {', '.join(f'"{col}" = excluded."{col}"' for col in df.columns)}
                    """
            
                    batch_data = []

                    for record in data:
                        new_record = []
                        for col, val in zip(df.columns, record):
                   
                            pg_type = type_mapping[col]
                            if val is None:
                                new_record.append(None)
                            elif pg_type == 'timestamp' and isinstance(val, np.datetime64):
                                new_record.append(pd.Timestamp(val).to_pydatetime().replace(tzinfo=None))

            
                            elif isinstance(val, datetime):
                                new_record.append(pd.Timestamp(val).to_pydatetime())
                            elif pg_type in ['timestamp', 'timestamp without time zone', 'timestamp with time zone'] and isinstance(val, np.datetime64):
                                new_record.append(pd.Timestamp(val).to_pydatetime().replace(tzinfo=None))  # Modified line
                            elif pg_type in ['double precision', 'real'] and not isinstance(val, str):
                                new_record.append(float(val))
                            elif isinstance(val, np.int64):  # Add this line to handle numpy.int64
                                new_record.append(int(val))
                            elif pg_type == 'integer' and not isinstance(val, int):
                                new_record.append(int(val))
                            else:
                                new_record.append(val)
                    
                        batch_data.append(new_record)

                        if len(batch_data) == batch_size:
                            try:
                                
                             
                                await connection.executemany(insert_query, batch_data)
                                batch_data.clear()
                            except Exception as e:
                                print(f"An error occurred while inserting the record: {e}")
                                await connection.execute('ROLLBACK')
                                raise

                    if batch_data:  # Last batch
                        try:
                            await connection.executemany(insert_query, batch_data)
                        except Exception as e:
                            print(f"An error occurred while inserting the last batch: {e}")
                            raise  # Transaction wil
    async def insert_webull_option(self, option_data):
            async with self.pool.acquire() as conn:
                
                await conn.execute("""
                    INSERT INTO webull_options (
                        open, high, low, close, change, change_ratio, strike, expiry,
                        call_put, oi, oi_change, active_level, cycle, volume,
                        latest_price_vol, delta, imp_vol, gamma, theta, vega, rho, option_symbol, underlying_symbol,trade_time, tickerid
                    ) VALUES (
                        $1, $2, $3, $4, $5, $6, $7, $8, $9, $10, $11, $12, $13, $14, $15, $16, 
                        $17, $18, $19, $20, $21, $22, $23, $24,$25
                    )
                    ON CONFLICT (option_symbol, underlying_symbol, strike, expiry, call_put)
                    DO UPDATE SET
                    open = EXCLUDED.open, 
                    high = EXCLUDED.high, 
                    low = EXCLUDED.low, 
                    close = EXCLUDED.close, 
                    change = EXCLUDED.change, 
                    change_ratio = EXCLUDED.change_ratio, 
                    oi = EXCLUDED.oi, 
                    oi_change = EXCLUDED.oi_change, 
                    active_level = EXCLUDED.active_level, 
                    cycle = EXCLUDED.cycle, 
                    volume = EXCLUDED.volume, 
                    latest_price_vol = EXCLUDED.latest_price_vol, 
                    delta = EXCLUDED.delta, 
                    imp_vol = EXCLUDED.imp_vol, 
                    gamma = EXCLUDED.gamma, 
                    theta = EXCLUDED.theta, 
                    vega = EXCLUDED.vega, 
                    rho = EXCLUDED.rho, 
                    trade_time = EXCLUDED.trade_time
                    """, 

                option_data['open'], option_data['high'], option_data['low'], option_data['close'],
                option_data['change'], option_data['change_ratio'], option_data['strike'], option_data['expiry'],
                option_data['call_put'], option_data['oi'], option_data['oi_change'], option_data['active_level'],
                option_data['cycle'], option_data['volume'], option_data['latest_price_vol'], option_data['delta'],
                option_data['imp_vol'], option_data['gamma'], option_data['theta'], option_data['vega'], 
                option_data['rho'], option_data['option_symbol'], option_data['underlying_symbol'],option_data['trade_time'], option_data['tickerid']
                )
    async def disconnect(self):
        await self.pool.close()

    async def fetch(self, query):
        async with self.pool.acquire() as conn:
            records = await conn.fetch(query)
            return records


import os
connection_string = os.environ.get('CONNECTION_STRING')
import aiohttp
import asyncio
import pandas as pd
from pytz import timezone
from .database_manager import DatabaseManager

from .webull_helpers import parse_most_active, parse_total_top_options, parse_contract_top_options, parse_ticker_values, parse_ipo_data, parse_etfs

class WebullMarkets(DatabaseManager):
    """General market data from webull"""
    def __init__(self, connection_string=None, pool=None):

        self.pool = pool
        self.connection_string = connection_string
        self.most_active_types = ['rvol10d', 'turnoverRatio', 'volume', 'range']
        self.top_option_types = ['totalVolume', 'totalPosition', 'volume', 'position', 'impVol', 'turnover']
        self.top_gainer_loser_types = ['afterMarket', 'preMarket', '5min', '1d', '5d', '1m', '3m', '52w']
        self.etf_types = ['industry', 'index', 'commodity', 'other']
        self.high_and_low_types = ['newHigh', 'newLow', 'nearHigh', 'nearLow']

       

        self.headers = {
        "App": "global",
        "App-Group": "broker",
        "Appid": "wb_web_app",
        "Device-Type": "Web",
        "Did": "8tb5au1228olpj2jss5vittmtk7pcvf6",
        "Hl": "en",
        "Locale": "eng",
        "Os": "web",
        "Osv": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/118.0.0.0 Safari/537.36",
        "Ph": "Windows Chrome",
        "Platform": "web",
        "Referer": "https://app.webull.com/",
        "Reqid": "a9d8d422e0e84041a035fb2389f18dae",
        "Sec-Ch-Ua": "\"Chromium\";v=\"118\", \"Google Chrome\";v=\"118\", \"Not=A?Brand\";v=\"99\"",
        "Sec-Ch-Ua-Mobile": "?0",
        "Sec-Ch-Ua-Platform": "\"Windows\"",
        "T_time": "1698276695206",
        "Tz": "America/Los_Angeles",
        "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/118.0.0.0 Safari/537.36",
        "Ver": "3.40.11",
        "X-S": "49ef20ad66d1e24a83ff8b2015bc13c6d133285c5665dbbe4aa6032572749931",
        "X-Sv": "xodp2vg9"
    }

    async def fetch_endpoint(self, endpoint):
        async with aiohttp.ClientSession(headers=self.headers) as session:
            async with session.get(endpoint) as resp:
                return await resp.json()
            
    

    async def get_top_options(self, rank_type:str='volume', as_dataframe:bool = True):
        """Rank Types:
        
        >>> totalVolume (total contract volume for a ticker)
        >>> totalPosition (total open interest for a ticker)
        >>> volume (volume by contract)
        >>> position (open interest by contract)
        >>> posIncrease (open interest increase)
        >>> posDecrease (open interest decrease)
        >>> impVol 
        >>> turnover

        DEFAULT: volume"""
        endpoint = f"https://quotes-gw.webullfintech.com/api/wlas/option/rank/list?regionId=6&rankType={rank_type}&pageIndex=1&pageSize=350"
        datas = await self.fetch_endpoint(endpoint)
        data = datas['data']
        if 'total' in rank_type:
            total_data = await parse_total_top_options(data)
            df = pd.DataFrame(total_data)
            
      
        else:
            total_data = await parse_contract_top_options(data)
            df= pd.DataFrame(total_data)
            
    
        if as_dataframe == False:
            return total_data
        
        df['rank_type'] = rank_type
        df.to_csv(f'data/top_options/top_options_{rank_type}.csv', index=False)
        df.columns = df.columns.str.lower()
        df = df.drop(columns=['dt', 'sectype', 'fatradetime', 'tradetime', 'status', 'template'])
        if self.connection_string is not None:
            df['totalasset'] = df['totalasset'].astype(float)
            df['netasset'] = df['netasset'].astype(float)
            df['implvol'] = df['implvol'].astype(float)
            df['position'] = df['position'].astype(float)
            df['middleprice'] = df['middleprice'].astype(float)
            df['turnover'] = df['turnover'].astype(float)
            df['positionchange'] = df['positionchange'].astype(float)
            df['unsymbol'] = df['unsymbol'].astype('string')
            df['strikeprice'] = df['strikeprice'].astype(float)
            df['price'] = df['price'].astype(float)
            df['direction'] = df['direction'].astype('string')
            # Convert columns to float
            float_columns = ['close', 'change', 'changeratio', 'marketvalue', 'volume', 'turnoverrate',
                            'pettm', 'preclose', 'fiftytwowkhigh', 'fiftytwowklow', 'open', 'high', 
                            'low', 'vibrateratio', 'pchratio', 'pprice', 'pchange']
            df[float_columns] = df[float_columns].astype(float)

    
          
     

  
 
            await self.connect()
            await self.batch_insert_dataframe(df, table_name=f'top_options_{rank_type.lower()}', unique_columns='insertion_timestamp')

        
        return df


    async def get_most_active(self, rank_type:str='rvol10d', as_dataframe:bool=False):
        """Rank types: 
        
        >>> volume
        >>> range
        >>> turnoverRatio
        >>> rvol10d
        
        """
        endpoint = f"https://quotes-gw.webullfintech.com/api/wlas/ranking/topActive?regionId=6&rankType={rank_type}&pageIndex=1&pageSize=350"
        datas = await self.fetch_endpoint(endpoint)
        parsed_data = await parse_most_active(datas)
        if as_dataframe == False:
            return parsed_data
        df = pd.DataFrame(parsed_data)
        df['rank_type'] = rank_type
        df.columns = df.columns.str.lower()
        if self.connection_string is not None:

            await self.connect()
            await self.batch_insert_dataframe(df, table_name=f'most_active_{rank_type}', unique_columns='insertion_timestamp')
        os.makedirs(f'data/top_active', exist_ok=True)

        df.to_csv(f'data/top_active/top_active_{rank_type}.csv', index=False)
        return df


        
    async def get_all_rank_types(self,types):
        """
        types:

        >>> wb.most_active_types
        >>> wb.top_option_types
        """
        if types == self.top_option_types:
            tasks = [self.get_top_options(type) for type in types]
            results = await asyncio.gather(*tasks)

            return results
        elif types == self.most_active_types:
            tasks = [self.get_most_active(type) for type in types]
            results = await asyncio.gather(*tasks)
            return results           


    async def earnings(self, start_date:str, pageSize: str='100', as_dataframe:str=True):
        """
        Pulls a list of earnings.

        >>> Start Date: enter a start date in YYYY-MM-DD format.

        >>> pageSize: enter the amount to be returned. default = 100

        >>> as_dataframe: default returns as a pandas dataframe.
        
        """
        endpoint = f"https://quotes-gw.webullfintech.com/api/bgw/explore/calendar/earnings?regionId=6&pageIndex=1&pageSize={pageSize}&startDate={start_date}"
        datas = await self.fetch_endpoint(endpoint)
        parsed_data = await parse_ticker_values(datas)
        if as_dataframe == False:
            return parsed_data
       
        df = pd.DataFrame(parsed_data)
        df.columns = df.columns.str.lower()
        df.to_csv('data/earnings/earnings_upcoming.csv', index=False)
        df = df.rename(columns={'v_releasedate': 'release_date'})
        if self.connection_string is not None:
            await self.connect()
            await self.batch_insert_dataframe(df, table_name='earnings', unique_columns='release_date')

        return df


    async def get_top_gainers(self, rank_type:str='1d', pageSize: str='100', as_dataframe:bool=True):
        """
        Rank Types:

        >>> afterMarket
        >>> preMarket
        >>> 5min
        >>> 1d (daily)
        >>> 5d (5day)
        >>> 1m (1month)
        >>> 3m (3month)
        >>> 52w (52 week)  

        DEFAULT: 1d (daily) 


        >>> PAGE SIZE:
            Number of results to return. Default = 100     
        """
        endpoint = f"https://quotes-gw.webullfintech.com/api/bgw/market/topGainers?regionId=6&rankType={rank_type}&pageIndex=1&pageSize={pageSize}"
        datas = await self.fetch_endpoint(endpoint)
        parsed_data = await parse_ticker_values(datas)
        if as_dataframe == False:
            return parsed_data
        df = pd.DataFrame(parsed_data)
        df['rank_type'] = rank_type
        df['gainer_type'] = 'topGainers'

        df.columns = df.columns.str.lower()
        df.to_csv(f'data/top_gainers/top_gainers_{rank_type}.csv', index=False)
        if self.connection_string is not None:
            if 't_sectype' in df.columns:
                df = df.drop(columns=['t_sectype'])
            await self.connect()
            await self.batch_insert_dataframe(df, table_name=f'top_gainers_{rank_type}', unique_columns='insertion_timestamp')
        return df
    

    async def get_top_losers(self, rank_type:str='1d', pageSize: str='100', as_dataframe:bool=True):
        """
        Rank Types:

        >>> afterMarket
        >>> preMarket
        >>> 5min
        >>> 1d (daily)
        >>> 5d (5day)
        >>> 1m (1month)
        >>> 3m (3month)
        >>> 52w (52 week)  

        DEFAULT: 1d (daily) 


        >>> PAGE SIZE:
            Number of results to return. Default = 100     
        """
        endpoint = f"https://quotes-gw.webullfintech.com/api/bgw/market/dropGainers?regionId=6&rankType={rank_type}&pageIndex=1&pageSize={pageSize}"
        datas = await self.fetch_endpoint(endpoint)
        parsed_data = await parse_ticker_values(datas)
        if as_dataframe == False:
            return parsed_data
        df = pd.DataFrame(parsed_data)
        df['rank_type'] = rank_type
        df['gainer_type'] = 'topLosers'

        df.to_csv(f'data/top_losers/top_losers{rank_type}.csv', index=False)
        df.columns = df.columns.str.lower()
        if self.connection_string is not None:
            if 't_sectype' in df.columns:
                df.drop(['t_sectype'], axis=1, inplace=True)
            await self.connect()
            await self.batch_insert_dataframe(df, table_name=f'top_losers_{rank_type}', unique_columns='insertion_timestamp')
        return df
    
    async def get_all_gainers_losers(self, type:str='gainers'):
        """TYPE OPTIONS:
        >>> gainers - all gainers across all rank_types
        >>> losers - all losers across all rank_types
        
        """
        types = self.top_gainer_loser_types
        if type == 'gainers':
            tasks = [self.get_top_gainers(type) for type in types]
            results = await asyncio.gather(*tasks)
            return results
        

        elif type == 'losers':
            tasks =[self.get_top_losers(type) for type in types]
            results = await asyncio.gather(*tasks)
            return results

    async def get_forex(self):
        endpoint = "https://quotes-gw.webullfintech.com/api/bgw/market/load-forex"
        datas = await self.fetch_endpoint(endpoint)

        df = pd.DataFrame(datas)

        df.to_csv('data/forex/forex_quotes.csv', index=False)
        df.columns = df.columns.str.lower()
        if self.connection_string is not None:
            if 't_sectype' in df.columns:
                df.drop(['t_sectype'], axis=1, inplace=True)
            await self.connect()
            await self.batch_insert_dataframe(df, table_name='forex', unique_columns='insertion_timestamp')
        return df
    
    async def etf_finder(self, type:str='industry'):
        """
        TYPES:

        >>> index
        >>> industry
        >>> commodity
        >>> other
        
        """
        endpoint = f"https://quotes-gw.webullfintech.com/api/wlas/etfinder/pcFinder?topNum=5&finderId=wlas.etfinder.{type}&nbboLevel=true"
        datas = await self.fetch_endpoint(endpoint)
        data = await parse_etfs(datas)

        df = pd.DataFrame(data)
        df['type'] = type
        df.to_csv(f'data/etfs/etfs_{type}.csv', index=False)
        df.columns = df.columns.str.lower()
        if self.connection_string is not None:
            df = df.drop(columns=['id', 'sectype', 'exchangetrade'])
            await self.connect()
            await self.batch_insert_dataframe(df, table_name=f'etfs_{type}', unique_columns='insertion_timestamp')

        return df
    
    async def get_all_etfs(self, types):
        types = self.etf_types
        tasks =[self.etf_finder(type) for type in types]

        results = await asyncio.gather(*tasks)

        return results


    async def highs_and_lows(self, type:str='newLow', pageSize:str='200', as_dataframe:bool=True):
        """
        TYPES:

        >>> newLow
        >>> newHigh
        >>> nearHigh
        >>> nearLow
        """
        endpoint = f"https://quotes-gw.webullfintech.com/api/wlas/ranking/52weeks?regionId=6&rankType={type}&pageIndex=1&pageSize={pageSize}"
        datas = await self.fetch_endpoint(endpoint)

        data = await parse_ticker_values(datas)

        if as_dataframe == False:
            return data
        
        df = pd.DataFrame(data)
        df.to_csv(f'data/highs_and_lows/{type}.csv', index=False)
        df.columns = df.columns.str.lower()
        return df
        
    async def ipos(self, type:str='filing', as_dataframe:bool=True):
        """
        TYPES:

        >>> filing
        >>> pricing
        
        """
        endpoint = f"https://quotes-gw.webullfintech.com/api/bgw/ipo/listIpo?regionId=6&status={type}&includeBanner=true"
        datas = await self.fetch_endpoint(endpoint)
        data = await parse_ipo_data(datas)

        if as_dataframe == False:
            return data
        
        df = pd.DataFrame(data)
        df.columns = df.columns.str.lower()
        df.to_csv(f'data/ipos/ipo_list_{type}.csv')

        return df
    

    


import sys
from pathlib import Path

# Add the project directory to the sys.path
project_dir = str(Path(__file__).resolve().parents[1])
if project_dir not in sys.path:
    sys.path.append(project_dir)


import aiohttp
import asyncio
import os

from fudstop.apis.misc.models.weather_models import WeatherLocation

API_KEY = os.environ.get('YOUR_WEATHER_KEY')



class Weather:
    def __init__(self):
        self.key = API_KEY



    async def get_coordinates(self, city_name, state, country_code, limit:str='2'):
        url = f"http://api.openweathermap.org/geo/1.0/direct?q={city_name},{state},{country_code}&limit={limit}&appid={API_KEY}"
        async with aiohttp.ClientSession() as session:
            async with session.get(url) as response:
                data = await response.json()
                if data:
                    return data[0]['lat'], data[0]['lon']
                else:
                    return None, None
                


    async def get_weather(self, city, state, country_code, limit:str='5', unit:str='F'):
        """
        Get the weather for the provided location.

        Example:

        >>> Dallas, TX, USA

        >>> boise, idaho, us

        >>> dallas texas, us

        >>> Sacramento, CA, usa

        Optional Args:


        >>> limit: the number of days of weather to return || default 5
        
        
        """

        lat, lon = await self.get_coordinates(city,state,country_code,limit)
        url = f"https://api.openweathermap.org/data/3.0/onecall?lat={lat}&lon={lon}&exclude=minutely,hourly,daily,alerts&appid={self.key}"
        async with aiohttp.ClientSession() as session:
            async with session.get(url) as resp:
                data = await resp.json()
                return WeatherLocation(data, unit=unit)





import sys
from pathlib import Path

# Add the project directory to the sys.path
project_dir = str(Path(__file__).resolve().parents[1])
if project_dir not in sys.path:
    sys.path.append(project_dir)


#import main libraries
import asyncio
from datetime import datetime
from collections import deque
from discord_webhook import AsyncDiscordWebhook, DiscordEmbed
from polygon.websocket import WebSocketClient, Market, Feed
from polygon.websocket import WebSocketClient


#import configuration, helpers, and conversion mappings
from market_handlers.cfg import hex_colors
from market_handlers.list_sets import CRYPTO_DESCRIPTIONS,CRYPTO_HOOKS
from fudstop.apis.polygonio.mapping import stock_condition_dict, option_condition_dict, OPTIONS_EXCHANGES


#import message handlers
from market_handlers.forex import handle_forex_msg
from market_handlers.stocks import handle_msg
from market_handlers.crypto import handle_crypto_msg
from market_handlers.options import handle_option_msg
#from market_handlers.indices import indices_handler

#import the most active tickers
from fudstop.list_sets.ticker_lists import most_active_tickers
#ensure no duplicates exist
most_active_tickers = set(most_active_tickers)


#define all markets
markets = [Market.Options, Market.Stocks, Market.Indices, Market.Crypto, Market.Forex]



#subscribe to all 5 polygon.io websocket clusters. [mix and match as needed]
subscription_patterns = {
    Market.Options: ["T.*,A.*"], #trades, aggregates
    Market.Stocks: ["A.*,T.*,Q.*"], #aggregates, trades, quotes
    #Market.Indices: ["AM.*"], # aggregates - per minute
    Market.Crypto: ['XT.*, XL2.*'], # crypto trades, level 2 book
    Market.Forex: ['CA.*, C.*'] # forex aggregates - per minute, forex quotes

}

# Mapping each market to its handler
market_handlers = {
    Market.Stocks: handle_msg,
    Market.Options: handle_option_msg,
    #Market.Indices: indices_handler,
    Market.Crypto: handle_crypto_msg,
    Market.Forex: handle_forex_msg
    }




# Initialize deques for EquityTrade and EquityOptionAgg
# to process batches of 250 incoming symbols / tickers 
# into postgres database


equity_trade_deque = deque(maxlen=250)
equity_option_agg_deque = deque(maxlen=250)


#import scripts
from fudstop.scripts.universal_snapshot import get_universal_snapshot



#import sdk kits for streamlined data analysis / processing
from fudstop.apis.discord_.discord_sdk import DiscordSDK
from fudstop.apis.stocksera_.stocksera_ import StockSera
from fudstop.apis.nasdaq.nasdaq_sdk import Nasdaq
from fudstop.apis.webull.webull_markets import WebullMarkets
from fudstop.apis.webull.webull_trading import WebullTrading
from fudstop.apis.polygonio.async_polygon_sdk import Polygon
from fudstop.apis.newyork_fed.newyork_fed_sdk import FedNewyork
from fudstop.apis.ofr.ofr_sdk import OFR



#set your API keys in an .env file. 
#see README.md


discord = DiscordSDK()
stockSera = StockSera()
nasdaq = Nasdaq()
wbmarkets = WebullMarkets(connection_string=None) #optional - pass a database connection_string
wbtrading = WebullTrading()
newyorkfed = FedNewyork()
ofr = OFR(connection_string=None) #optional - pass a database connection_string




#the consumer function for data processing in real time

async def consumer(queue: asyncio.Queue):
    batch = []

    while True:
        data = await queue.get()
        #print(queue.qsize()) debugging line 
      
        


        type = data.get('type')








        """
        >>> WORK ZONE...



        PERFORM ALL ANALYSIS AND DATABASE INTEGRATION HERE.


        THIS IS REAL-TIME DATA PROCESSING.


        YOU HAVE SEVERAL TOOLS AT YOUR DISPOSAL VIA THE VARIOUS APIS.


        THE SKY IS YOUR LIMIT.
        
        """







        #define and perform conversions for
        #stock T.* (EquityTrade)

        if type == 'EquityTrade':
            stock_symbol = data.get('stock_symbol')
            exchange = data.get('exchange')
            price = data.get('trade_price')
            size = data.get('trade_size')
            conditions = data.get('trade_conditions')
            timestamp = data.get('trade_timestamp')
            for condition in conditions:
                if condition in stock_condition_dict and size > 1000:
               
                    hook = stock_condition_dict[condition]
                  
                
                    if condition not in stock_condition_dict:
                        print(f'New condition / unsaved condition found: {condition}')

                    


        #defining the EquityAgg
        #stock instance variables
        #A.*



        elif type == 'EquityAgg':
            
            stock_symbol = data.get('stock_symbol')
            if stock_symbol in most_active_tickers:
                close = data.get('close_price')
                high = data.get('high_price')
                open = data.get('open_price')
                low = data.get('low_price')
                volume = data.get('volume')
                official_open = data.get('official_open')
                total_volume = data.get('accumulated_volume')
                vwap = data.get('vwap_price')
                timestamp = data.get('timestamp')


                #.. do work here .. #




        #defining the EquityQuote
        #stock instance variables
        #Q.*

        elif type == 'EquityQuote':
            
            quote_symbol = data.get('quote_symbol')
            if quote_symbol in most_active_tickers:
                ask = data.get('ask')
                bid = data.get('bid')
                ask_size = data.get('ask_size')
                bid_size = data.get('bid_size')
                ask_exchange = data.get('ask_exchange')
                bid_exchange = data.get('bid_exchange')
                timestamp = data.get('timestamp')
                tape = data.get('tape')


                # do work here #





















        #define the EquityOptionAggregates instance

        """
        >>> EXAMPLE WITH LIVE FEATURE ENGINEERING EXAMPLE
        

        !. Several features have been engineered to display versatility in application on top of the provided attributes
        from the polygon API.


        >>> EXAMPLE WITH LIVE DATABASE INTEGRATION EXAMPLE

        !. As real-time option aggregates hit the SIP feed, they are batched in chunks of 250, passed to the polygon.io UniversalSnapshot
        API, and the real-time data and latest reference data are combined & stored in database in real-time.


        """


        if type =='EquityOptionAgg':

            option_symbol = data.get('option_symbol')
            if option_symbol is not None:
               
                equity_option_agg_deque.append(option_symbol)
                
                if len(batch) >= 250:
                    collected_aggs_symbols = ','.join(equity_option_agg_deque)
                    try:
                        data = await get_universal_snapshot(collected_aggs_symbols)


                    
                
                        

                        

                        for i in data:
                            trade_timestamp = i.get('last_trade.sip_timestamp')
                            if trade_timestamp is not None:
                                trade_timestamp = datetime.fromtimestamp(trade_timestamp / 1e9)

                            trade_conditions = i.get('last_trade.conditions', None)
                            if trade_conditions is not None:
                                trade_conditions = option_condition_dict.get(trade_conditions[0])

                            data_dict = {
                                'name': i.get('name'),
                                'option_symbol': i.get('ticker'),
                                'underlying_symbol': i.get('underlying_asset.ticker'),
                                'strike': i.get('details.strike_price'),
                                'call_put': i.get('details.contract_type'),
                                'expiry': (i.get('details.expiration_date')),
                                'underlying_price': i.get('underlying_asset.price'),
                                'change': i.get('session.change'),
                                'change_percent': i.get('session.change_percent'),
                                'early_change': i.get('session.early_trading_change'),
                                'early_change_percent': i.get('session.early_trading_change_percent'),
                                'change_to_break_even': i.get('underlying_asset.change_to_break_even'),
                                'break_even_price': i.get('break_even_price'),
                                'open': i.get('session.open'),
                                'high': i.get('session.high'),
                                'low': i.get('session.low'),
                                'close': i.get('session.close'),
                                'previous_close': i.get('session.previous_close'),
                                'volume': i.get('session.volume'),
                                'oi': i.get('open_interest'),
                                'iv': i.get('implied_volatility'),
                                'delta': i.get('greeks.delta'),
                                'gamma': i.get('greeks.gamma'),
                                'theta': i.get('greeks.theta'),
                                'vega': i.get('greeks.vega'),
                                'trade_size': i.get('last_trade.size'),
                                'trade_price': i.get('last_trade.price'),
                                'trade_exchange': OPTIONS_EXCHANGES.get(i.get('last_trade.exchange')),
                                'trade_conditions': trade_conditions,
                                'trade_timestamp': trade_timestamp,
                                'ask': i.get('last_quote.ask'),
                                'ask_size': i.get('last_quote.ask_size'),
                                'ask_exchange': OPTIONS_EXCHANGES.get(i.get('last_quote.ask_exchange')),
                                'bid': i.get('last_quote.bid'),
                                'bid_size': i.get('last_quote.bid_size'),
                                'bid_exchange': OPTIONS_EXCHANGES.get(i.get('last_quote.bid_exchange'))}
                                                



                            if data_dict['expiry'] is not None:
                                current_date = datetime.now()

                                def parse_date(date_str):
                                    year, month, day = map(int, date_str.split('-'))
                                    return datetime(year, month, day)
                                expiry_date = parse_date(data_dict['expiry'])
                            if data_dict['underlying_price'] is not None and data_dict['strike'] is not None and data_dict['strike'] != 0:
                                data_dict['moneyness'] = data_dict['underlying_price'] / data_dict['strike']
                            else:
                                data_dict['moneyness'] = None

                            # Calculate Time to Expiry (in days)
                            if data_dict['expiry'] is not None:
                                try:
                                    expiry_date = datetime.strptime(data_dict['expiry'], '%Y-%m-%d')
                                    current_date = datetime.now()
                                    if expiry_date is not None and current_date is not None:
                                        data_dict['time_to_expiry'] = (expiry_date - current_date).days
                                    else:
                                        data_dict['time_to_expiry'] = None
                                except ValueError:
                                    # Handle date parsing errors if needed
                                    data_dict['time_to_expiry'] = None
                            else:
                                data_dict['time_to_expiry'] = None
                            # Calculate Intrinsic Value

                            # Calculate Intrinsic Value
                            if data_dict.get('underlying_price') is not None and data_dict.get('strike') is not None:
                                data_dict['intrinsic_value_call'] = max(0, data_dict['underlying_price'] - data_dict['strike'])
                                data_dict['intrinsic_value_put'] = max(0, data_dict['strike'] - data_dict['underlying_price'])
                            else:
                                data_dict['intrinsic_value_call'] = None
                                data_dict['intrinsic_value_put'] = None

                            # Calculate Extrinsic Value
                            if all(k in data_dict and data_dict[k] is not None for k in ['close', 'intrinsic_value_call', 'intrinsic_value_put']):
                                data_dict['extrinsic_value'] = data_dict['close'] - max(data_dict['intrinsic_value_call'], data_dict['intrinsic_value_put'])
                            else:
                                data_dict['extrinsic_value'] = None



                            # Calculate Liquidity Score
                            if data_dict.get('volume') is not None and data_dict.get('oi') is not None:
                                data_dict['liquidity_score'] = data_dict['volume'] * data_dict['oi']
                            else:
                                data_dict['liquidity_score'] = None

                            # Calculate Implied Leverage
                            if data_dict.get('underlying_price') is not None and data_dict.get('close') is not None and data_dict['close'] != 0:
                                data_dict['implied_leverage'] = data_dict['underlying_price'] / data_dict['close']
                            else:
                                data_dict['implied_leverage'] = None

                            # Calculate Delta to Theta Ratio
                            if data_dict.get('delta') is not None and data_dict.get('theta') is not None and data_dict['theta'] != 0:
                                data_dict['delta_to_theta_ratio'] = data_dict['delta'] / data_dict['theta']
                            else:
                                data_dict['delta_to_theta_ratio'] = None

                            # Calculate Cost of Theta using Close Price
                            if data_dict.get('theta') is not None and data_dict.get('close') is not None and data_dict['close'] != 0:
                                data_dict['cost_of_theta'] = data_dict['theta'] / data_dict['close']
                            else:
                                data_dict['cost_of_theta'] = None

                            # # Calculate Risk-Reward Ratio
                            # data_dict['risk_reward_ratio'] = data_dict['potential_profit'] / data_dict['potential_loss']

    
                            batch.append(data_dict)

                            if len(batch) >= 250:

                                await asyncio.sleep(900)
                                batch.clear()
                                        
                    except Exception as e:
                        print(f'Error {e}')


        """
        EXAMPLE WITH  DISCORD USING LIVE INDICES DATA


        >>> *** Tutorial Coming Soon *** <<<       
        
        """


        #!. Create a couple of channels in your discord server, create a webhook for each. One for SPX and one for NDX related indices.
        spx_indices = ""
        ndx_indices = ""
        #spx_indices = os.environ.get('spx_indices')
        #ndx_indices = os.environ.get('ndx_indices')








        if type == 'Indices':
            ticker = data.get('ticker')
            min_open = data.get('minute_open')
            min_high = data.get('minute_high')
            min_low = data.get('minute_low')
            min_close = data.get('minute_close')
            start = data.get('minute_start')
            end = data.get('minute_end')
            official_open = data.get('day_open')
            name = data.get('name')


            #if 'SPX' in name:
                #!. Uncomment if you have a webhook 
                #hook = AsyncDiscordWebhook(spx_indices)
                #embed = DiscordEmbed(title=f"INDICES MARKET - SPX RELATED", description=f'```py\n{name}```', color=hex_colors['magenta'])
                #embed.add_embed_field(name=name, value=f"{ticker}")
                #embed.add_embed_field(name=f"Day Stats:", value=f"> Open: **{data.get('day_open')}**\n> Now: **{data.get('minute_close')}**")
                #embed.add_embed_field(name=f"1min Aggs:", value=f"> Open: **${round(float(min_open),2)}**\n> High: **${round(float(min_high),2)}**\n> Low: **${round(float(min_low),2)}**\n> Close: **${round(float(min_close),2)}**")
                #embed.add_embed_field(name=f"Recorded Time:", value=f"> Start: **{start}**\n> End: **{end}**")
                #hook.add_embed(embed)
                #await hook.execute()

            #if 'NASDAQ' in name:
                #!. Uncomment if you have a webhook 
                #hook = AsyncDiscordWebhook(ndx_indices)
                #embed = DiscordEmbed(title=f"INDICES MARKET - SPX RELATED", description=f'```py\n{name}```', color=hex_colors['magenta'])
                #embed.add_embed_field(name=name, value=f"{ticker}")
                #embed.add_embed_field(name=f"Day Stats:", value=f"> Open: **{data.get('day_open')}**\n> Now: **{data.get('minute_close')}**")
                #embed.add_embed_field(name=f"1min Aggs:", value=f"> Open: **${round(float(min_open),2)}**\n> High: **${round(float(min_high),2)}**\n> Low: **${round(float(min_low),2)}**\n> Close: **${round(float(min_close),2)}**")
                #embed.add_embed_field(name=f"Recorded Time:", value=f"> Start: **{start}**\n> End: **{end}**")
                #hook.add_embed(embed)
                #await hook.execute()





        #defining OptionTrades - realtime options!


        if type == 'EquityOptionTrade':
            expiry = data.get('expiry')
            option_symbol = data.get('option_symbol')
            call_put = data.get('call_put')
            strike = data.get('strike')
            underlying_symbol = data.get('underlying_symbol')
            option_symbol = data.get('option_symbol')
            price = data.get('price')
            price_change = data.get('price_change')
            size = data.get('size')
            volume_change = data.get('volume_change')
            conditions = data.get('conditions')
            exchange = data.get('exchange')
            price_to_strike = data.get('price_to_strike')
            hour_of_day = data.get('hour_of_day')
            weekday = data.get('weekday')
            timestamp = data.get('timestamp')
            dollar_cost = (100 * price) * size
        










        #defining crypto trades! real-time!

        if type == 'XT':
            if data.get('conditions') == 'Buy Side':
                color = hex_colors['green']


            elif data.get('conditions') == 'Sell Side':
                color = hex_colors['red']


            #using a dictionary of crypto hooks
            if data.get('symbol') in CRYPTO_HOOKS:
                price = data.get('price')
                size = data.get('size')
                hook = CRYPTO_HOOKS[data.get('symbol')]
                desc = CRYPTO_DESCRIPTIONS.get(data.get('symbol'))
                
                webhook = AsyncDiscordWebhook(hook, content="@everyone <@375862240601047070>")
                embed = DiscordEmbed(title=f"{data.get('symbol')} | Live Trades", description=f"```py\n{desc}```", color=color)
                embed.add_embed_field(name=f"Exchange:", value=f"> **{data.get('exchange')}**")
                embed.add_embed_field(name=f"Side:", value=f"> **{data.get('conditions')}**")
                embed.add_embed_field(name="Trade Info:", value=f"> Price: **${price}**\n> Size: **{size}**")
                embed.add_embed_field(name=f"Time:", value=f"> **{data.get('timestamp')}**")

                embed.set_footer(text='Data by Polygon.io | Implemented by FUDSTOP | https://www.discord.gg/fudstop')
                embed.set_timestamp()

                webhook.add_embed(embed)

                asyncio.gather(send_webhooks(webhook))


async def send_webhooks(webhook):
    await webhook.execute()


import logging

import os

logging.basicConfig(level=logging.INFO)

import asyncio



#run it all!

#if you have postgres.. uncomment the code below and provide your credentials.



#more detailed tutorial coming!


#this program leverages all 5 real-time markets from polygon.io and several SDK kits to perform deep, real-time data analysis across all major markets

#discord and postgres integrated

async def main():
    num_workers = int(os.environ.get("NUM_WORKERS", 50))
    #max_connections = int(os.environ.get("MAX_CONNECTIONS", 10))
    #connection_semaphore = asyncio.Semaphore(max_connections)

    logging.info("Starting up...")
    #await db.connect()
    data_queue = asyncio.Queue()  # Set a reasonable size limit

    # Create consumer tasks
    consumer_tasks = [asyncio.create_task(consumer(data_queue)) for _ in range(num_workers)]

    # Create WebSocket client tasks
    websocket_tasks = []
    for market in markets:
        #async with connection_semaphore:
        client = WebSocketClient(api_key=os.environ.get('YOUR_POLYGON_KEY'), subscriptions=subscription_patterns[market], market=market, feed=Feed.RealTime)
        websocket_tasks.append(client.connect(lambda msgs, handler=market_handlers[market]: handler(msgs, data_queue)))

    try:
        await asyncio.gather(*websocket_tasks)
        await asyncio.gather(*consumer_tasks)
    except Exception as e:
        logging.error(f"An error occurred: {e}")
        # Implement some error recovery logic here

asyncio.run(main())


import sys
from pathlib import Path

# Add the project directory to the sys.path
project_dir = str(Path(__file__).resolve().parents[2])
if project_dir not in sys.path:
    sys.path.append(project_dir)

import os
from dotenv import load_dotenv
# Assuming we have the array of ticker IDs as numpy arrays from the user's environment
import numpy as np
from fudstop.apis.webull.webull_helpers import convert_to_date
from aiohttp.client_exceptions import ContentTypeError
from fudstop.apis.webull.trade_models.options import WebullOptionsData, VolumeAnalysis
import aiohttp
import asyncio
import json
import asyncpg
load_dotenv()
GEX_KEY = os.environ.get('GEXBOT')

print(GEX_KEY)

from list_sets.ticker_lists import most_active_tickers
most_active_tickers = set(most_active_tickers)
from fudstop.apis.webull.webull_trading import WebullTrading
from datetime import datetime, timedelta

trading = WebullTrading()

class WebullOptions:
    def __init__(self, connection_string = None):
        self.pool = None
        self.connection_string = connection_string
        self.api_key = os.environ.get('YOUR_POLYGON_KEY')
        self.today = datetime.now().strftime('%Y-%m-%d')
        self.yesterday = (datetime.now() - timedelta(days=1)).strftime('%Y-%m-%d')
        self.tomorrow = (datetime.now() + timedelta(days=1)).strftime('%Y-%m-%d')
        self.thirty_days_ago = (datetime.now() - timedelta(days=30)).strftime('%Y-%m-%d')
        self.thirty_days_from_now = (datetime.now() + timedelta(days=30)).strftime('%Y-%m-%d')
        self.fifteen_days_ago = (datetime.now() - timedelta(days=15)).strftime('%Y-%m-%d')
        self.fifteen_days_from_now = (datetime.now() + timedelta(days=15)).strftime('%Y-%m-%d')
        self.eight_days_from_now = (datetime.now() + timedelta(days=8)).strftime('%Y-%m-%d')
        self.eight_days_ago = (datetime.now() - timedelta(days=8)).strftime('%Y-%m-%d')
        self.headers = {
        "Access_token": "dc_us_tech1.18ba751f1ae-05a653bff5ec4b10a2da53487bd4dfb6",
        "Accept": "*/*",
        "App": "global",
        "App-Group": "broker",
        "Appid": "wb_web_app",
        "Content-Type": "application/json",
        "Device-Type": "Web",
        "Did": "8tb5au1228olpj2jss5vittmtk7pcvf6",
        "Hl": "en",
        "Locale": "eng",
        "Os": "web",
        "Osv": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/118.0.0.0 Safari/537.36",
        "Ph": "Windows Chrome",
        "Platform": "web",
        "Referer": "https://app.webull.com/",
        "Reqid": "a9d8d422e0e84041a035fb2389f18dae",
        "Sec-Ch-Ua": "\"Chromium\";v=\"118\", \"Google Chrome\";v=\"118\", \"Not=A?Brand\";v=\"99\"",
        "Sec-Ch-Ua-Mobile": "?0",
        "Sec-Ch-Ua-Platform": "\"Windows\"",
        "T_time": "1698276695206",
        "Tz": "America/Los_Angeles",
        "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/118.0.0.0 Safari/537.36",
        "Ver": "3.40.11",
        "X-S": "49ef20ad66d1e24a83ff8b2015bc13c6d133285c5665dbbe4aa6032572749931",
        "X-Sv": "xodp2vg9"
    }

        if connection_string is not None:
            connection_string = os.environ.get('WEBULL_OPTIONS')
    async def connect(self):
        self.pool = await asyncpg.create_pool(
            dsn=self.connection_string, min_size=1, max_size=10
        )

    async def fetch(self, query, *args):
        """
        Fetch data from the database using the provided SQL query.

        :param query: The SQL query to execute.
        :param args: The arguments to pass to the SQL query.
        :return: The result of the query as a list of records.
        """
        async with self.pool.acquire() as conn:  # Acquire a connection from the pool
            # Execute the query with the provided arguments
            return await conn.fetch(query, *args)


    async def batch_insert_options(self, pairs):
        try:
            async with self.pool.acquire() as conn:  # Acquire a connection from the pool

                async with conn.transaction():  # Start a transaction
                    # Prepare the statement to insert data
                    insert_query = 'INSERT INTO webull_options (symbol, ticker_id, ticker) VALUES ($1, $2, $3)'
                    # Perform the batch insert
                    await conn.executemany(insert_query, pairs)
                    print("Batch insert completed.")
        except asyncpg.exceptions.UniqueViolationError:
            print(f'Duplicate found - skipping.')


    async def yield_batch_ids(self, ticker_symbol):
   
        async with self.pool.acquire() as conn:
            # We will fetch all derivative IDs associated with the ticker symbol
            derivative_ids = await conn.fetch(
                'SELECT ticker_id FROM webull_options WHERE ticker = $1',
                ticker_symbol
            )
            
            # Convert the records to a list of IDs
            derivative_id_list = [str(record['ticker_id']) for record in derivative_ids]

            # Yield batches of 55 IDs at a time as a comma-separated string
            for i in range(0, len(derivative_id_list), 55):
                yield ','.join(derivative_id_list[i:i+55])

    async def get_option_ids(self, ticker):
        ticker_id = await trading.get_ticker_id(ticker)
        params = {
            "tickerId": f"{ticker_id}",
            "count": -1,
            "direction": "all",
            "expireCycle": [1,
                3,
                2,
                4
            ],
            "type": 0,
            "quoteMultiplier": 100,
            "unSymbol": f"{ticker}"
        }
        data = json.dumps(params)
        url="https://quotes-gw.webullfintech.com/api/quote/option/strategy/list"

        # Headers you may need to include, like authentication tokens, etc.
        headers = trading.headers
        # The body of your POST request as a Python dictionary
        import pandas as pd
        # Make the POST request
        # Make the POST request
        async with aiohttp.ClientSession(headers=headers) as session:
            async with session.post(url, data=data) as resp:
                response_json = await resp.json()
                print(response_json)
                # Extract the 'expireDateList' from the response
                expireDateList = response_json.get('expireDateList')

                # Flatten the nested 'data' from each item in 'expireDateList'
            try:
                data_flat = [item for sublist in expireDateList if sublist and sublist.get('data') for item in sublist['data']]



                # Create a DataFrame from the flattened data
                df_cleaned = pd.DataFrame(data_flat)

                # Drop the 'askList' and 'bidList' columns if they exist
                df_cleaned.drop(columns=['askList', 'bidList'], errors='ignore', inplace=True)
                # Existing DataFrame columns
                df_columns = df_cleaned.columns

                # Original list of columns you want to convert to numeric
                numeric_cols = ['open', 'high', 'low', 'strikePrice', 'isStdSettle', 'quoteMultiplier', 'quoteLotSize']

                # Filter the list to include only columns that exist in the DataFrame
                existing_numeric_cols = [col for col in numeric_cols if col in df_columns]

                # Now apply the to_numeric conversion only to the existing columns
                df_cleaned[existing_numeric_cols] = df_cleaned[existing_numeric_cols].apply(pd.to_numeric, errors='coerce')

      
                print(df_cleaned)
                df_cleaned.to_csv('test.csv', index=False)


                # Load the data from the CSV file
                df = pd.read_csv('test.csv')

                # Extract 'tickerId' column values in batches of 55
                ticker_ids = df['tickerId'].unique()  # Assuming 'tickerId' is a column in your DataFrame
                symbol_list = df['symbol'].unique().tolist()
            # Pair up 'tickerId' and 'symbol'
                # Before you call batch_insert_options, make sure pairs contain the correct types
                pairs = [(str(symbol), int(ticker_id), str(ticker)) for ticker_id, symbol in zip(ticker_ids, symbol_list)]

                
                await self.batch_insert_options(pairs)
               
            except (ContentTypeError, TypeError):
                print(f'Error for {ticker}')
    async def get_option_id_for_symbol(self, ticker_symbol):
        async with self.pool.acquire() as conn:
            # Start a transaction
            async with conn.transaction():
                # Execute the query to get the option_id for a given ticker_symbol
                # This assumes 'symbol' column exists in 'options_data' table and 
                # is used to store the ticker symbol
                query = f'''
                    SELECT ticker_id FROM webull_options
                    WHERE ticker = '{ticker_symbol}';
                '''
                # Fetch the result
                result = await conn.fetch(query)
                # Return a list of option_ids or an empty list if none were found
                return [record['ticker_id'] for record in result]


    async def get_option_symbols_by_ticker_id(self, ticker_id):
        async with self.pool.acquire() as conn:
            # Start a transaction
            async with conn.transaction():
                # Execute the query to get all option_symbols for a given ticker_id
                query = '''
                    SELECT option_symbol FROM options_data
                    WHERE ticker_id = $1;
                '''
                # Fetch the result
                records = await conn.fetch(query, ticker_id)
                # Extract option_symbols from the records
                return [record['option_symbol'] for record in records]
    async def get_ticker_symbol_pairs(self):
        # Assume 'pool' is an instance variable pointing to a connection pool
        async with self.pool.acquire() as conn:
            # Start a transaction
            async with conn.transaction():
                # Create a cursor for iteration using 'cursor()' instead of 'execute()'
                async for record in conn.cursor('SELECT ticker_id, symbol FROM webull_options'):
                    yield (record['ticker_id'], record['symbol'])

    async def get_volume_analysis(self, ticker):
        ticker_id = await trading.get_ticker_id(ticker)
        params = {
            "tickerId": f"{ticker_id}",
            "count": -1,
            "direction": "all",
            "expireCycle": [1,
                3,
                2,
                4
            ],
            "type": 0,
            "quoteMultiplier": 100,
            "unSymbol": f"{ticker}"
        }
        data = json.dumps(params)
        url="https://quotes-gw.webullfintech.com/api/quote/option/strategy/list"

        # Headers you may need to include, like authentication tokens, etc.
        headers = trading.headers
        # The body of your POST request as a Python dictionary
        import pandas as pd
        # Make the POST request
        # Make the POST request
        async with aiohttp.ClientSession(headers=headers) as session:
            async with session.post(url, data=data) as resp:
                response_json = await resp.json()
          
                # Extract the 'expireDateList' from the response
                expireDateList = response_json.get('expireDateList')

                # Flatten the nested 'data' from each item in 'expireDateList'
            try:
                data_flat = [item for sublist in expireDateList if sublist and sublist.get('data') for item in sublist['data']]



                # Create a DataFrame from the flattened data
                df_cleaned = pd.DataFrame(data_flat)

                # Drop the 'askList' and 'bidList' columns if they exist
                df_cleaned.drop(columns=['askList', 'bidList'], errors='ignore', inplace=True)

                # Convert specified columns to numeric values, coercing errors to NaN
                numeric_cols = ['open', 'high', 'low', 'strikePrice', 'isStdSettle', 'quoteMultiplier', 'quoteLotSize']
                # Iterate through the list of numeric columns and check if they exist in df_cleaned
                existing_numeric_cols = [col for col in numeric_cols if col in df_cleaned.columns]

                # Now apply the conversion only on the columns that exist
                df_cleaned[existing_numeric_cols] = df_cleaned[existing_numeric_cols].apply(pd.to_numeric, errors='coerce')

                print(df_cleaned)
                df_cleaned.to_csv('test.csv', index=False)


                # Load the data from the CSV file
                df = pd.read_csv('test.csv')

                # Extract 'tickerId' column values in batches of 55
                ticker_ids = df['tickerId'].unique()  # Assuming 'tickerId' is a column in your DataFrame
                symbol_list = df['symbol'].unique().tolist()
            # Pair up 'tickerId' and 'symbol'
                pairs = list(zip(ticker_ids, symbol_list))

                
                # Split into batches of 55
                batches = [ticker_ids[i:i + 55] for i in range(0, len(ticker_ids), 55)]

                ticker_id_strings = [','.join(map(str, batch)) for batch in batches]







                for ticker_id_string in ticker_id_strings:
                    ticker_ids = ticker_id_string.split(',')
                    for deriv_id in ticker_ids:
                        all_data = []
                        volume_analysis_url = f"https://quotes-gw.webullfintech.com/api/statistic/option/queryVolumeAnalysis?count=200&tickerId={deriv_id}"
                        async with aiohttp.ClientSession(headers=headers) as session:
                            async with session.get(volume_analysis_url) as resp:
                                data = await resp.json()
                                all_data.append(data)


                   
                        return all_data
                        #df = pd.DataFrame(all_data)
                        #df.to_csv('all_options', index=False)
            except (ContentTypeError, TypeError):
                print(f'Error for {ticker}')

    async def get_option_ids_limited(self, sem, ticker):
        async with sem:
            # This will wait until the semaphore allows entry (i.e., under the limit)
            return await self.get_option_ids(ticker)

    async def harvest_options(self,most_active_tickers):
        # Set the maximum number of concurrent requests
        max_concurrent_requests = 5  # For example, limiting to 10 concurrent requests

        # Create a semaphore with your desired number of concurrent requests
        sem = asyncio.Semaphore(max_concurrent_requests)
        await self.connect()
        # Create tasks using the semaphore
        tasks = [self.get_option_ids_limited(sem, ticker) for ticker in most_active_tickers]

        # Run the tasks concurrently and wait for all to complete
        await asyncio.gather(*tasks)


    async def get_option_data_for_ticker(self, ticker):
        print(f"Starting processing for ticker: {ticker}")
        dataframes = []  # Initialize a list to collect DataFrames

        async for info in self.yield_batch_ids(ticker_symbol=ticker):
            print(f"Processing batch ID: {info} for ticker: {ticker}")
            url = f"https://quotes-gw.webullfintech.com/api/quote/option/quotes/queryBatch?derivativeIds={info}"
            async with aiohttp.ClientSession(headers=trading.headers) as session:
                async with session.get(url) as resp:
                    data = await resp.json()
                    if not data:  # If data is empty or None, break the loop
                        print(f"No more data for ticker: {ticker}. Moving to next.")
                        break
                    wb_data = WebullOptionsData(data)
                    df = wb_data.as_dataframe
                    df['ticker'] = ticker
                    await self.insert_dataframe_in_batches(df, 'options_data')
            
        print(f"Finished processing for ticker: {ticker}")

  



    async def insert_dataframe_in_batches(self, df, table_name, batch_size=55):
        """
        Insert a pandas DataFrame into a SQL table in batches.

        :param df: The pandas DataFrame to insert.
        :param table_name: The name of the target SQL table.
        :param batch_size: The size of the batches to insert.
        """
        # Make sure we have a connection
        
        df['expiry_date'] = df['expiry_date'].apply(convert_to_date)
        df['close'] = df['close'].astype(float)
        df['open'] = df['open'].astype(float)
        df['high'] = df['high'].astype(float)
        df['low'] = df['low'].astype(float)
        df['open_interest'] = df['open_interest'].replace({np.nan: None})
        async with self.pool.acquire() as conn:
            try:
                # Convert DataFrame to list of tuples
                records = df.to_records(index=False)
                columns = df.columns.tolist()
                values = [tuple(x) for x in records]

                # Create a prepared statement
                placeholders = ', '.join(f'${i+1}' for i in range(len(columns)))
                insert_query = f'INSERT INTO {table_name} ({", ".join(columns)}) VALUES ({placeholders})'

                # Insert data in batches
                for i in range(0, len(values), batch_size):
                    batch = values[i:i + batch_size]
                    await conn.executemany(insert_query, batch)

            except asyncpg.exceptions.UniqueViolationError:
                print(f'Duplicate found - skipping.')


    # Initialize an HTTP session
    async def fetch_volume_analysis(self, ticker_id):
        volume_analysis_url = f"https://quotes-gw.webullfintech.com/api/statistic/option/queryVolumeAnalysis?count=500&tickerId={ticker_id}"
        async with aiohttp.ClientSession(headers=self.headers) as session:
            async with session.get(volume_analysis_url) as resp:
                data = await resp.json()
                if data is not None:
                    return VolumeAnalysis(data)
                
    # Initialize an HTTP session
    async def test_vol_anal(self, ticker_id):
        volume_analysis_url = f"https://quotes-gw.webullfintech.com/api/statistic/option/queryVolumeAnalysis?count=500&tickerId={ticker_id}"
        async with aiohttp.ClientSession(headers=self.headers) as session:
            async with session.get(volume_analysis_url) as resp:
                data = await resp.json()
                if data is not None:
                    return VolumeAnalysis(data).trades_and_dates_dict
    async def insert_trades_and_dates(self, data):
        data['date'] = datetime.strptime(data['date'], '%Y-%m-%d')
        # Convert numeric strings to appropriate numeric types
        data['price'] = float(data['price']) if data['price'] else 0.0
        data['buy'] = int(data['buy']) if data['buy'] else 0
        data['sell'] = int(data['sell']) if data['sell'] else 0
        data['volume'] = int(data['volume']) if data['volume'] else 0
        data['strike_price'] = float(data['strike_price']) if data['strike_price'] else 0.0
        # Convert expiry_date string to date object
        data['expiry_date'] = datetime.strptime(data['expiry_date'], '%Y-%m-%d')

                

        # Assuming 'data' is a dictionary containing the keys that match your table's columns.
        async with self.pool.acquire() as connection:
            await connection.execute('''
                INSERT INTO trades_and_dates (
                    date, price, buy, sell, volume, ratio,
                    option_symbol, symbol, strike_price, call_put,
                    expiry_date
                ) VALUES($1, $2, $3, $4, $5, $6, $7, $8, $9, $10, $11)
            ''',
            data['date'], data['price'], data['buy'], data['sell'], data['volume'],
            data['ratio'], data['option_symbol'], data['symbol'], data['strike_price'],
            data['call_put'], data['expiry_date'])
    async def insert_volume_analysis_data(self, data: dict):
        # Assuming data contains the fields corresponding to your self.data_dict
        # You would extract these values and insert them into your volume_analysis table
        async with self.pool.acquire() as conn:
            await conn.execute("""
                INSERT INTO volume_analysis (ticker_id, symbol, option_id, total_trades, total_volume, avg_price, buy_volume, sell_volume, neutral_volume, option_symbol, strike_price, call_put, expiry_date)
                VALUES ($1, $2, $3, $4, $5, $6, $7, $8, $9, $10, $11, $12, $13)
            """, data.get('ticker_id'), data.get('symbol', None), data.get('option_id', None), data.get('total_trades', None), data.get('total_volume', None), data.get('avg_price',None),data.get('buy_volume',None), data.get('sell_volume', None), data.get('neutral_volume', None), data.get('option_symbol', None), data.get('strike_price') , data.get('call_put'), data.get('expiry_date')        )
        

    async def store_options_data(self):
        await self.connect()
        semaphore = asyncio.Semaphore(5)  # Adjust the number to limit concurrent tasks

        async def limited_get_option_data_for_ticker(ticker):
            async with semaphore:
                return await self.get_option_data_for_ticker(ticker)

        tasks = [limited_get_option_data_for_ticker(i) for i in most_active_tickers]
        await asyncio.gather(*tasks)

    async def close_pool(self):
        await self.pool.close()



import sys
from pathlib import Path

# Add the project directory to the sys.path
project_dir = str(Path(__file__).resolve().parents[1])
if project_dir not in sys.path:
    sys.path.append(project_dir)

import os
import stocksera
import pandas as pd

from typing import List
from concurrent.futures import ThreadPoolExecutor, as_completed
from .models.daily_treasury import DailyTreasuryData
from .models.sec_filings import SECFilingsData
from .models.market_news import MarketNewsData
from .models.news_sentiment import NewsSentimentData
from .models.insider_trades import InsiderTrades
from .models.ftds import FailureToDeliverData
from .models.low_float import LowFloatData
from .models.highest_shorted import HighestShortedData
from .models.short_volume import ShortVolumeData
from .models.inflation import InflationData
from .models.borrowed_shares import BorrowedSharesData
from .models.jobless_claims import JoblessClaimsData
from .models.jim_cramer import JimCramerData
from .models.retail_sales import RetailSalesData
from .models.reverse_repo import ReverseRepoData
from datetime import datetime, timedelta
from dotenv import load_dotenv
load_dotenv()

YOUR_STOCKSERA_KEY = os.environ.get('YOUR_STOCKSERA_KEY')
print(YOUR_STOCKSERA_KEY)
class StockSera:
    def __init__(self):


        self.client = stocksera.Client(YOUR_STOCKSERA_KEY)


        self.today = datetime.now().strftime('%Y-%m-%d')
        self.yesterday = (datetime.now() - timedelta(days=1)).strftime('%Y-%m-%d')
        self.tomorrow = (datetime.now() + timedelta(days=1)).strftime('%Y-%m-%d')
        self.thirty_days_ago = (datetime.now() - timedelta(days=30)).strftime('%Y-%m-%d')
        self.thirty_days_from_now = (datetime.now() + timedelta(days=30)).strftime('%Y-%m-%d')
        self.fifteen_days_ago = (datetime.now() - timedelta(days=15)).strftime('%Y-%m-%d')
        self.fifteen_days_from_now = (datetime.now() + timedelta(days=15)).strftime('%Y-%m-%d')
        self.eight_days_from_now = (datetime.now() + timedelta(days=8)).strftime('%Y-%m-%d')
        self.eight_days_ago = (datetime.now() - timedelta(days=8)).strftime('%Y-%m-%d')


    def borrowed_shares(self, ticker_or_tickers, concurrency=None) -> List[BorrowedSharesData]:
        def fetch_data(ticker):
            data = self.client.borrowed_shares(ticker)
            df = pd.DataFrame(data)
            df['ticker'] = df['ticker'].astype(str)
            df['fee'] = df['fee'].astype(float)
            df['available'] = df['available'].astype(int)
            df['date_updated'] = pd.to_datetime(df['date_updated'])
            df.columns = df.columns.str.lower()
            return df

        if concurrency and isinstance(ticker_or_tickers, list):
            data_frames = []
            with ThreadPoolExecutor(max_workers=concurrency) as executor:
                futures = {executor.submit(fetch_data, ticker): ticker for ticker in ticker_or_tickers}

                for future in as_completed(futures):
                    ticker = futures[future]
                    try:
                        data = future.result()
                    except Exception as exc:
                        print(f"{ticker} generated an exception: {exc}")
                    else:
                        print(data)
                        data_frames.append(data)

            combined_df = pd.concat(data_frames, ignore_index=True)
            return combined_df

        else:
            data = self.client.borrowed_shares(ticker_or_tickers)
            df = pd.DataFrame(data)
            df['ticker'] = df['ticker'].astype(str)
            df['fee'] = df['fee'].astype(float)
            df['available'] = df['available'].astype(int)
            df['date_updated'] = pd.to_datetime(df['date_updated'])
            df.columns = df.columns.str.lower()
            df.to_csv(f'data/stocksera/borrowed_shares_{ticker_or_tickers}.csv', index=False)  # Corrected the variable here
            return df
        

    def short_volume(self, ticker_or_tickers, date_from='2019-09-17', date_to='2023-10-27', concurrency=None) -> List[ShortVolumeData]:
        """
        REQUIRED:

        >>> ticker



        OPTIONAL: 
        >>> date_from: the date to start surveying
        >>> date_to: the date to stop surveying

        >>> concurrency: Run this command on a list of tickers rather than a single ticker.
        """
        
        def fetch_data(ticker):
            data = self.client.short_volume(ticker, date_from, date_to)
            df = pd.DataFrame(data)
            df.columns = df.columns.str.lower()
            df = df.rename({'short vol': 'short_vol', 'short exempt vol': 'short_exempt_vol', 'total vol': 'total_vol', '% shorted': 'percent_shorted'})
            df['date'] = pd.to_datetime(df['date'])
            df['ticker'] = ticker
            return df
        
        if concurrency and isinstance(ticker_or_tickers, list):
            data_frames = []
            
            with ThreadPoolExecutor(max_workers=concurrency) as executor:
                futures = {executor.submit(fetch_data, ticker): ticker for ticker in ticker_or_tickers}
                
                for future in as_completed(futures):
                    ticker = futures[future]
                    try:
                        data = future.result()
                    except Exception as exc:
                        print(f"{ticker} generated an exception: {exc}")
                    else:
                        data_frames.append(data)

            if data_frames:  # Check if list is not empty
                combined_df = pd.concat(data_frames, ignore_index=True)

                combined_df.columns = combined_df.columns.str.lower()
                print(combined_df.columns)
                combined_df = combined_df.rename({'short vol': 'short_vol', 'short exempt vol': 'short_exempt_vol', 'total vol': 'total_vol', '% shorted': 'percent_shorted'})
                combined_df['date'] = pd.to_datetime(combined_df['date'])

                return combined_df
            else:
                print("No data to concatenate.")
                return None
                
        else:
            return fetch_data(ticker_or_tickers)

    def daily_treasury(self, days: str = '100') -> List[DailyTreasuryData]:
        """
        Arguments:

        OPTIONAL:

        >>> days: number of days to survey
        
        """
        data = self.client.daily_treasury(days)
        df = pd.DataFrame(data)

        # Rename columns to standard underscore style
        df.columns = df.columns.str.lower().str.replace(" ", "_").str.replace("%", "percent").str.replace("?", "question")

        all_treasury_data = []
        for index, row in df.iterrows():
            daily_treasury_data = DailyTreasuryData(
                date=row['date'],
                close_balance=row['close_balance'],
                open_balance=row['open_balance'],
                amount_change=row['amount_change'],
                percent_change=row['percent_change'],
                moving_avg=row['moving_avg']
            )

            all_treasury_data.append(daily_treasury_data)

        return all_treasury_data
    


    def failure_to_deliver(self, ticker_or_tickers, date_from='2023-10-28', date_to='2024-01-01', concurrency=None) -> List[FailureToDeliverData]:
        """
        Arguments:

        required:

        >>> ticker or a list of tickers


        OPTIONAL:

        >>> concurrency: add concurrency when passing in a list
        
        """
        def fetch_data(ticker):
            data = self.client.ftd(ticker, date_from, date_to)
            df = pd.DataFrame(data)
            df['ticker'] = ticker  # Assuming the API doesn't return the ticker, add it manually
            df['date'] = pd.to_datetime(df['date'])
            df['ftd'] = df['ftd'].astype(float)
            df['price'] = df['price'].astype(float)
            df['ftd_x_$'] = df['ftd_x_$'].astype(float)
            df['t+35_date'] = pd.to_datetime(df['t+35_date'])
            df.columns = df.columns.str.lower()
            return df

        if concurrency and isinstance(ticker_or_tickers, list):
            data_frames = []
            with ThreadPoolExecutor(max_workers=concurrency) as executor:
                futures = {executor.submit(fetch_data, ticker): ticker for ticker in ticker_or_tickers}

                for future in as_completed(futures):
                    ticker = futures[future]
                    try:
                        data = future.result()
                    except Exception as exc:
                        print(f"{ticker} generated an exception: {exc}")
                    else:
                        data_frames.append(data)

            combined_df = pd.concat(data_frames, ignore_index=True)
            return combined_df

        else:
            data = self.client.ftd(ticker_or_tickers, date_from, date_to)
            df = pd.DataFrame(data)
            df['ticker'] = ticker_or_tickers  # Assuming the API doesn't return the ticker, add it manually
            df['date'] = pd.to_datetime(df['date'])
            df['ftd'] = df['ftd'].astype(float)
            df['price'] = df['price'].astype(float)
            df['ftd_x_$'] = df['ftd_x_$'].astype(float)
            df['t+35_date'] = pd.to_datetime(df['t+35_date'])
            df.columns = df.columns.str.lower()
            df.to_csv(f'data/stocksera/failure_to_deliver_{ticker_or_tickers}.csv', index=False)
            return df

    def highest_shorted(self) -> List[HighestShortedData]:
        """
        Arguments

        >>> None


        """

        data = self.client.short_interest()

        df = pd.DataFrame(data)
        df.columns = df.columns.str.lower()
      
        df.columns = df.columns.str.lower().str.replace(" ", "_").str.replace("%", "percent_")
        short_interest = []
        for i, row in df.iterrows():
            short_interest_data = HighestShortedData(rank=row['rank'], 
                                                     ticker=row['ticker'], 
                                                     date=row['date'], 
                                                     short_interest=row['short_interest'], 
                                                     average_volume= row['average_volume'],
                                                     days_to_cover=row['days_to_cover'],
                                                     percent_float_short=row['percent_float_short'])

            short_interest.append(short_interest_data)

        return short_interest
    def inflation(self, year:str=None):
        """
        Arguments:
        >>> year: OPTIONAL - the year to survey (default all results)
        """

        data = self.client.inflation()
        if year is None:
            data = InflationData(data)
            df = pd.DataFrame(data)
            df.columns = df.columns.str.lower()
            return df
        else:
            data = InflationData(data).get_inflation_by_year(year)
            df = pd.DataFrame(data, index=[year])
            df.columns = df.columns.str.lower()
            return df



    def jobless_claims(self, days:str='100', as_dataframe:bool=True) -> List[JoblessClaimsData]:
        """
        Arguments:

        >>> days: the number of days to survey (optional - default 100)
        >>> as_dataframe: return as a dataframe (optional -default False)
        """
        

        data = self.client.jobless_claims(days)
        formatted_data = [{k.lower().replace(' ', '_'): v for k, v in item.items()} for item in data]
        if as_dataframe == True:
            
            jobless_data_dicts = [JoblessClaimsData(**item).as_dict() for item in formatted_data]
       
            df = pd.DataFrame(jobless_data_dicts)
            df.columns = df.columns.str.lower()
            df = df[::-1]
            return df

        data = [JoblessClaimsData(**item) for item in formatted_data]

        return data
        



    
    def insider_trading(self, as_dataframe:bool=True) -> List[InsiderTrades]:
        """
        Arguments:

        >>> as_dataframe: optional - returns as a pandas dataframe. (default True)
        """
        
        data = self.client.insider_trading()
        formatted_data = [{k.lower().replace(' ', '_'): v for k, v in item.items()} for item in data]
        if as_dataframe == False:
            formatted_data = [InsiderTrades(**i) for i in formatted_data]
            return formatted_data

        df = pd.DataFrame(formatted_data)
        return df


  
    def jim_cramer(self, as_dataframe:bool=True) -> List[JimCramerData]:
        """
        Arguments:
        >>> as_dataframe: optional - returns as a pandas dataframe (default True)
        """

        data = self.client.jim_cramer()
        formatted_data = [{k.lower().replace(' ', '_'): v for k, v in item.items()} for item in data]
        if as_dataframe == False:
            formatted_data = [JimCramerData(**i) for i in formatted_data]
            return formatted_data
        df = pd.DataFrame(formatted_data)
        return df
       

   


    def low_float(self, as_dataframe: bool= True):
        """
        Arguments:

        >>> as_dataframe: optional - returns as a pandas dataframe (default True)
        """
        
        data = self.client.low_float()
        formatted_data = [{k.lower().replace(' ', '_'): v for k, v in item.items()} for item in data]

        if as_dataframe == False:
            formatted_data = LowFloatData(formatted_data)
            return formatted_data

        df = pd.DataFrame(formatted_data)
        return df
       



    def sec_filings(self, ticker_or_tickers, concurrency:str=None):
        """
        Arguments:

        >>> ticker_or_tickers: pass in a single ticker or list of tickers

        optional:

        >>> concurrency: if passing in a list - set concurrency levels
        """
        def fetch_data(ticker):
            data = self.client.sec_fillings(ticker)
            data = [{k.lower().replace(' ', '_'): v for k, v in item.items()} for item in data]
            df = pd.DataFrame(data)
            
            df['ticker'] = ticker
            df['filling'] = df['filling'].astype(str)
            df['description'] = df['description'].astype(str)
            df['filling_date'] = pd.to_datetime(df['filling_date'])
            df.columns = df.columns.str.lower()
            return df

        if concurrency and isinstance(ticker_or_tickers, list):
            data_frames = []
            with ThreadPoolExecutor(max_workers=concurrency) as executor:
                futures = {executor.submit(fetch_data, ticker): ticker for ticker in ticker_or_tickers}

                for future in as_completed(futures):
                    ticker = futures[future]
                    try:
                        data = future.result()
                    except Exception as exc:
                        print(f"{ticker} generated an exception: {exc}")
                    else:
                        data_frames.append(data)

            combined_df = pd.concat(data_frames, ignore_index=True)
            return combined_df

        else:
            data = self.client.sec_fillings(ticker_or_tickers)
            df = pd.DataFrame(data)
            df['ticker'] = ticker_or_tickers  # Assuming the API doesn't return the ticker, add it manually
            df['filling'] = df['filling'].astype(str)
            df['description'] = df['description'].astype(str)
            df['filling_date'] = pd.to_datetime(df['filling_date'])
            df.columns = df.columns.str.lower()
            df.to_csv(f'data/stocksera/sec_filings_{ticker_or_tickers}.csv', index=False)
            return df



 
    def news_sentiment(self, ticker_or_tickers, concurrency:str=None):
        """
        Arguments:

        required: 
        >>> ticker_or_tickers: pass in either a single ticker or list of tickers.

        optional:
        >>> if passing in a list of tickers, set concurrency level as needed.
        
        """
        def fetch_data(ticker):
            data = self.client.news_sentiment(ticker)
            df = pd.DataFrame(data)
            df.columns = df.columns.str.lower()
            df['ticker'] = ticker  # Assuming the API doesn't return the ticker, add it manually
        
            return df

        if concurrency and isinstance(ticker_or_tickers, list):
            data_frames = []
            with ThreadPoolExecutor(max_workers=concurrency) as executor:
                futures = {executor.submit(fetch_data, ticker): ticker for ticker in ticker_or_tickers}

                for future in as_completed(futures):
                    ticker = futures[future]
                    try:
                        data = future.result()
                        
                    except Exception as exc:
                        print(f"{ticker} generated an exception: {exc}")
                    else:
                        data_frames.append(data)

            combined_df = pd.concat(data_frames, ignore_index=True)
            return combined_df

        else:
            data = self.client.news_sentiment(ticker_or_tickers)
            df = pd.DataFrame(data)
            print(df.columns)
            df['ticker'] = ticker_or_tickers  # Assuming the API doesn't return the ticker, add it manually
            df['date'] = pd.to_datetime(df['date'])
            df['title'] = df['title'].astype(str)
            df['link'] = df['link'].astype(str)
            df['sentiment'] = df['sentiment'].astype(str)
            df.columns = df.columns.str.lower()
            df.to_csv(f'data/stocksera/news_sentiment_{ticker_or_tickers}.csv', index=False)
            return df
   


    def market_news(self, as_dataframe:bool=True) -> List[MarketNewsData]:
        """
        Arguments:

        >>> as_dataframe: optional - returns as a pandas dataframe (default True)
        
        """

        data = self.client.market_news()
        formatted_data = [{k.lower().replace(' ', '_'): v for k, v in item.items()} for item in data]

        if as_dataframe == False:
            formatted_data = [MarketNewsData(**i) for i in formatted_data]
            return formatted_data

        df = pd.DataFrame(formatted_data)
        

        return df


    def retail_sales(self, days:str='100', as_dataframe:bool=True) -> List[RetailSalesData]:

        """
        Arguments:

        >>> days: optional: the days to survey (default 100)
        >>> as_dataframe: optional - returns as a pandas dataframe (default true).
        """

        data = self.client.retail_sales(days)
        formatted_data = [{k.lower().replace(' ', '_'): v for k, v in item.items()} for item in data]
        if as_dataframe == False:
            formatted_data = [RetailSalesData(**i) for i in formatted_data]
            return formatted_data

        df = pd.DataFrame(formatted_data)
        df = df[::-1] #reverse the order
        return df
       



    def reverse_repo(self, days:str='100', as_dataframe: bool=True) -> List[ReverseRepoData]:
        """
        Arguments:
        >>> days: optional - the number of days to survey
        >>> as_dataframe: optional - returns as a pandas dataframe (default True)
        """

        data = self.client.reverse_repo(days)
        formatted_data = [{k.lower().replace(' ', '_'): v for k, v in item.items()} for item in data]

        if as_dataframe == False:
            formatted_data = [ReverseRepoData(**i) for i in formatted_data]
            return formatted_data

        df = pd.DataFrame(formatted_data)


        return df


hex_color_dict = {
    "red": "FF0000",
    "green": "00FF00",
    "blue": "0000FF",
    "yellow": "FFFF00",
    "purple": "800080",
    "cyan": "00FFFF",
    "magenta": "FF00FF",
    "lime": "00FF00",
    "pink": "FFC0CB",
    "teal": "008080",
    "lavender": "E6E6FA",
    "brown": "A52A2A",
    "beige": "F5F5DC",
    "maroon": "800000",
    "mint": "98FF98",
    "olive": "808000",
    "coral": "FF7F50",
    "navy": "000080",
    "grey": "808080",
    "white": "FFFFFF",
    "black": "000000",
    "orange": "FFA500",
    "indigo": "4B0082",
    "violet": "EE82EE",
    "gold": "FFD700",
    "silver": "C0C0C0",
    "rose": "FF007F",
    "emerald": "50C878",
    "ruby": "E0115F",
    "sapphire": "0F52BA",
    "amber": "FFBF00",
    "crimson": "DC143C",
    "darkgreen": "006400",
    "darkblue": "00008B",
    "darkred": "8B0000",
    "lightgreen": "90EE90",
    "lightblue": "ADD8E6",
    "lightyellow": "FFFFE0",
    "lightpink": "FFB6C1",
    "darkpurple": "800080",
    "lightpurple": "D8BFD8",
    "turquoise": "40E0D0",
    "tan": "D2B48C",
    "lightgrey": "D3D3D3",
    "darkgrey": "A9A9A9",
    "ivory": "FFFFF0",
    "khaki": "F0E68C",
    "lavenderblush": "FFF0F5",
    "lightcoral": "F08080",
    "salmon": "FA8072",
    "darksalmon": "E9967A",
    "lightsalmon": "FFA07A",
    "darkorange": "FF8C00",
    "linen": "FAF0E6",
    "peachpuff": "FFDAB9",
    "seashell": "FFF5EE",
    "snow": "FFFAFA",
    "wheat": "F5DEB3",
    "mistyrose": "FFE4E1",
    "azure": "F0FFFF",
    "aliceblue": "F0F8FF",
    "honeydew": "F0FFF0",
    "mintcream": "F5FFFA"
}


import sys
from pathlib import Path
# Add the project directory to the sys.path
project_dir = str(Path(__file__).resolve().parents[1])
if project_dir not in sys.path:
    sys.path.append(project_dir)

from typing import List, Optional
from dataclasses import dataclass, field
from polygon_helpers import convert_datetime_list
from polygonio.mapping import stock_condition_dict,STOCK_EXCHANGES
from typing import Dict
from datetime import datetime

@dataclass
class Ticker:


    ticker: Optional[str] = None
    today_change: Optional[float] = None
    today_change_perc: Optional[float] = None


@dataclass
class Day:


    open: Optional[float] = None
    high: Optional[float] = None
    low: Optional[float] = None
    close: Optional[float] = None
    volume: Optional[int] = None
    vwap: Optional[float] = None


@dataclass
class LastQuote:


    ask_price: Optional[float] = None
    ask_size: Optional[float] = None
    bid_price: Optional[float] = None
    bid_size: Optional[float] = None
    quote_timestamp: Optional[float] = None


@dataclass
class LastTrade:


    conditions: Optional[List[int]] = None
    trade_id: Optional[str] = None
    trade_price: Optional[float] = None
    trade_size: Optional[int] = None
    trade_timestamp: Optional[int] = None
    trade_exchange: Optional[int] = None


@dataclass
class Min:


    accumulated_volume: Optional[int] = None
    minute_timestamp: Optional[float] = None
    vwap: Optional[float] = None
    volume: Optional[float] = None
    open: Optional[float] = None
    high: Optional[float] = None
    low: Optional[float] = None
    close: Optional[float] = None


@dataclass
class PrevDay:


    open: Optional[float] = None
    high: Optional[float] = None
    low: Optional[float] = None
    close: Optional[float] = None
    volume: Optional[float] = None
    vwap: Optional[float] = None


@dataclass
class StockSnapshot():  # Inheritance
    def __init__(self):
        super().__init__()

        
    ticker: Optional[str] = None
    today_changep: Optional[float] = None
    today_change: Optional[float] = 0.0
    stock_day: Optional[Day] = None
    stock_last_quote: Optional[LastQuote] = None
    last_trade: Optional[LastTrade] = None
    stock_minute_bar: Optional[Min] = None
    prev_day: Optional[PrevDay] = None

    
    def __init__(self, ticker_data):
        self.ticker = ticker_data.get('ticker', None)
        self.today_changep = ticker_data.get('todaysChangePerc', None)
        self.today_change = ticker_data.get('todaysChange', 0.0)

        day_data = ticker_data.get('day', {})
        if day_data:
            self.stock_day = Day(
                open=day_data.get('o'),
                high=day_data.get('h'),
                low=day_data.get('l'),
                close=day_data.get('c'),
                volume=day_data.get('v'),
                vwap=day_data.get('vw'),
            )

        quote_data = ticker_data.get('lastQuote', None)
        if quote_data:
            self.stock_last_quote = LastQuote(
                ask_price=quote_data.get('P'),
                ask_size=quote_data.get('S'),
                bid_price=quote_data.get('p'),
                bid_size=quote_data.get('s'),
                quote_timestamp=quote_data.get('t'),
            )
        else:
            quote_data = None
        trade_data = ticker_data.get('lastTrade', None)
        if trade_data:
            self.last_trade = LastTrade(
                conditions = [stock_condition_dict.get(i) for i in trade_data.get('c')],

                trade_id=trade_data.get('i'),
                trade_price=trade_data.get('p'),
                trade_size=trade_data.get('s'),
                timestamp = convert_datetime_list(trade_data.get('t'), unit='ns'),
                exchange = STOCK_EXCHANGES.get(trade_data.get('x'))

        
            )
        else:
            trade_data = None
        min_data = ticker_data.get('min', {})
        if min_data:
            self.stock_minute_bar = Min(
                accumulated_volume=min_data.get('av'),
                minute_timestamp=min_data.get('t'),
                vwap=min_data.get('vw'),
                volume=min_data.get('v'),
                open=min_data.get('o'),
                high=min_data.get('h'),
                low=min_data.get('l'),
                close=min_data.get('c'),
            )
        else:
            min_data = None
        prev_data = ticker_data.get('prevDay', None)
        if prev_data:
            self.prev_day = PrevDay(
                open=prev_data.get('o'),
                high=prev_data.get('h'),
                low=prev_data.get('l'),
                close=prev_data.get('c'),
                volume=prev_data.get('v'),
                vwap=prev_data.get('vw'),
            )
        else:
            prev_data = None





import pandas as pd


class AuctionResult:
    def __init__(self, auctions):
        self.auctionStatus = [auction.get('auctionStatus') for auction in auctions]
        self.operationId = [auction.get('operationId') for auction in auctions]
        self.operationDate = [auction.get('operationDate') for auction in auctions]
        self.operationType = [auction.get('operationType') for auction in auctions]
        self.operationDirection = [auction.get('operationDirection') for auction in auctions]
        self.method = [auction.get('method') for auction in auctions]
        self.releaseTime = [auction.get('releaseTime') for auction in auctions]
        self.closeTime = [auction.get('closeTime') for auction in auctions]
        self.classType = [auction.get('classType') for auction in auctions]
        self.note = [auction.get('note') for auction in auctions]
        self.totalSubmittedOrigFace = [auction.get('totalSubmittedOrigFace') for auction in auctions]
        self.totalAcceptedOrigFace = [auction.get('totalAcceptedOrigFace') for auction in auctions]
        self.totalSubmittedCurrFace = [auction.get('totalSubmittedCurrFace') for auction in auctions]
        self.totalAcceptedCurrFace = [auction.get('totalAcceptedCurrFace') for auction in auctions]
        self.totalAmtSubmittedPar = [auction.get('totalAmtSubmittedPar') for auction in auctions]
        self.totalAmtAcceptedPar = [auction.get('totalAmtAcceptedPar') for auction in auctions]
        self.settlementDate = [auction.get('settlementDate') for auction in auctions]
        self.lastUpdated = [auction.get('lastUpdated') for auction in auctions]

    @classmethod
    def from_dict(cls, data_dict):
        return cls(**data_dict)
    


class FXSwaps:
    def __init__(self, operations):
        self.operationType = [operation.get("operationType", []) for operation in operations]
        self.counterparty = [operation.get("counterparty", []) for operation in operations]
        self.currency = [operation.get("currency", []) for operation in operations]
        self.tradeDate = [operation.get("tradeDate", []) for operation in operations]
        self.settlementDate = [operation.get("settlementDate", []) for operation in operations]
        self.maturityDate = [operation.get("maturityDate", []) for operation in operations]
        self.termInDays = [operation.get("termInDays", []) for operation in operations]
        self.amount = [operation.get("amount", []) for operation in operations]
        self.interestRate = [operation.get("interestRate", []) for operation in operations]
        self.isSmallValue = [operation.get("isSmallValue", []) for operation in operations]
        self.lastUpdated = [operation.get("lastUpdated", []) for operation in operations]

        self.dict = {
            'operationType': self.operationType,
            'counterparty': self.counterparty,
            'currency': self.currency,
            'tradeDate': self.tradeDate,
            'settlementDate': self.settlementDate,
            'maturityDate': self.maturityDate,
            'termInDays': self.termInDays,
            'amount': self.amount,
            'interestRate': self.interestRate,
            'isSmallValue': self.isSmallValue,
            'lastUpdated': self.lastUpdated
        }
        

        self.df = pd.DataFrame(self.dict)


class TimeSeries:
    def __init__(self, timeseries):

        self.seriesbreak = [i['seriesbreak'] if i['seriesbreak'] is not None else None for i in timeseries]
        self.keyid = [i['keyid'] if i['keyid'] is not None else None for i in timeseries]
        self.description = [i['description'] if i['description'] is not None else None for i in timeseries]

        self.dict = { 
            'Description': self.description,
            'Key ID': self.keyid,
            'Series Break': self.seriesbreak}
        

        self.df = pd.DataFrame(self.dict)



class TimeSeriesData:
    def __init__(self, timeseries):
        self.asofdate = [i['asofdate'] if 'asofdate' in i else None for i in timeseries]
        self.keyid = [i['keyid'] if 'keyid' in i else None for i in timeseries]
        self.value = [i['value'] if 'value' in i  else None for i in timeseries]

        self.dict = { 
            'As of Date': self.asofdate,
            'Key ID': self.keyid,
            'Value': self.value}
        

        self.df = pd.DataFrame(self.dict)

class AsOfDates:
    def __init__(self, timeseries):

        self.seriesbreak = [i['seriesbreak'] if i['seriesbreak'] is not None else None for i in timeseries]
        self.asof = [i['asof'] if i['asof'] is not None else None for i in timeseries]

        self.dict = { 
            'Series Break': self.seriesbreak,
            'As of': self.asof,}

        

        self.df = pd.DataFrame(self.dict)



class SecuredReferenceRates:
    def __init__(self, refrates):

        self.effectiveDate = [i['effectiveDate'] if i['effectiveDate'] is not None else None for i in refrates]
        self.rate_type= [i['type'] if i['type'] is not None else None for i in refrates]
        self.average30day= [i['average30day'] if 'average30day' in i else None for i in refrates]
        self.average90day= [i['average90day'] if 'average90day' in i else None for i in refrates]
        self.average180day= [i['average180day'] if 'average180day' in i else None for i in refrates]
        self.index= [i['index'] if 'index' in i else None for i in refrates]
        self.revisionIndicator= [i['revisionIndicator'] if i['revisionIndicator'] is not None else None for i in refrates]
        self.percent = [i["percent"] if 'percent' in i else None for i in refrates]
        self.percentPercentile1 = [i["percentPercentile1"] if "percentPercentile1" in i else None for i in refrates]
        self.percentPercentile25 = [i["percentPercentile25"] if "percentPercentile25" in i  else None for i in refrates]
        self.percentPercentile75 = [i["percentPercentile75"] if "percentPercentile75" in i  else None for i in refrates]
        self.percentPercentile99 = [i["percentPercentile99"] if "percentPercentile99" in i  else None for i in refrates]
        self.TargetRateFrom = [i["targetRateFrom"] if "targetRateFrom" in i  else None for i in refrates]
        self.TargetRateTo = [i["targetRateTo"] if "targetRateTo" in i else None for i in refrates]
        self.VolumeInBillions = [i["volumeInBillions"] if "volumeInBillions" in i  else None for i in refrates]
        self.FootNoteID = [i["footnoteId"] if "footnoteId" in i else None for i in refrates]

        self.data_dict = { 

            'Effective Date': self.effectiveDate,
            'Rate Type': self.rate_type,
            'Average 30 Day': self.average30day,
            'Average 90 Day': self.average90day,
            'Average 180 Day': self.average180day,
            'Index': self.index,
            'Revision Indicator': self.revisionIndicator,
            "Rate Type": self.rate_type,
            "Percent": self.percent,
            "1st Percentile": self.percentPercentile1,
            "25th Percentile": self.percentPercentile25,
            "75th Percentile": self.percentPercentile75,
            "99th Percentile": self.percentPercentile99,
            "revisionIndicator": self.revisionIndicator,
            "Target Rate From": self.TargetRateFrom,
            "Target Rate To": self.TargetRateTo,
            "Volume In Billions": self.VolumeInBillions,
            "FootNote ID": self.FootNoteID
        }


        self.df = pd.DataFrame(self.data_dict)



class RepoOperations:
    def __init__(self, operations):
        self.operationId = [i['operationId'] if i['operationId'] is not None else None for i in operations]
        self.auctionStatus = [i['auctionStatus'] if 'auctionStatus' in i else None for i in operations]
        self.operationDate = [i['operationDate'] if 'operationDate' in i else None for i in operations]
        self.settlementDate = [i['settlementDate'] if 'settlementDate' in i else None for i in operations]
        self.maturityDate = [i['maturityDate'] if 'maturityDate' in i else None for i in operations]
        self.operationType = [i['operationType'] if 'operationType' in i else None for i in operations]
        self.operationMethod = [i['operationMethod'] if 'operationMethod' in i else None for i in operations]
        self.settlementType = [i['settlementType'] if 'settlementType' in i else None for i in operations]
        self.termCalenderDays = [i['termCalenderDays'] if 'termCalenderDays' in i else None for i in operations]
        self.term = [i['term'] if 'term' in i else None for i in operations]
        self.releaseTime = [i['releaseTime'] if 'releaseTime' in i else None for i in operations]
        self.closeTime = [i['closeTime'] if 'closeTime' in i else None for i in operations]
        self.note = [i['note'] if 'note' in i else None for i in operations]
        self.lastUpdated = [i['lastUpdated'] if 'lastUpdated' in i else None for i in operations]
        self.participatingCpty = [i['participatingCpty'] if 'participatingCpty' in i else None for i in operations]
        self.acceptedCpty = [i['acceptedCpty'] if 'acceptedCpty' in i else None for i in operations]
        self.totalAmtSubmitted = [i['totalAmtSubmitted'] if 'totalAmtSubmitted' in i else None for i in operations]
        self.totalAmtAccepted = [i['totalAmtAccepted'] if 'totalAmtAccepted' in i else None for i in operations]
        details = [i['details'] if 'details' in i else None for i in operations]
        indexed_details = [(i, detail) for i, detail in enumerate(details)]
        second_detail = indexed_details[1][1]
        self.securityType = [i['securityType'] if 'securityType' in i else None for i in second_detail]
        self.amtSubmitted = [i['amtSubmitted'] if 'amtSubmitted' in i else None for i in second_detail]
        self.amtAccepted = [i['amtAccepted'] if 'amtAccepted' in i else None for i in second_detail]
        self.minimumBidRate = [i['minimumBidRate'] if 'minimumBidRate' in i else None for i in second_detail]
        self.percentHighRate = [i['percentHighRate'] if 'percentHighRate' in i else None for i in second_detail]
        self.percentLowRate = [i['percentLowRate'] if 'percentLowRate' in i else None for i in second_detail]
        self.percentStopOutRate = [i['percentStopOutRate'] if 'percentStopOutRate' in i else None for i in second_detail]
        self.percentWeightedAverageRat = [i['percentWeightedAverageRat'] if 'percentWeightedAverageRat' in i else None for i in second_detail]

        # Create the data dictionary
        self.data_dict = {
            'operationId': self.operationId,
            'auctionStatus': self.auctionStatus,
            'operationDate': self.operationDate,
            'settlementDate': self.settlementDate,
            'maturityDate': self.maturityDate,
            'operationType': self.operationType,
            'operationMethod': self.operationMethod,
            'settlementType': self.settlementType,
            'termCalenderDays': self.termCalenderDays,
            'term': self.term,
            'releaseTime': self.releaseTime,
            'closeTime': self.closeTime,
            'note': self.note,
            'lastUpdated': self.lastUpdated,
            'participatingCpty': self.participatingCpty,
            'acceptedCpty': self.acceptedCpty,
            'totalAmtSubmitted': self.totalAmtSubmitted,
            'totalAmtAccepted': self.totalAmtAccepted,
}

        # Create a dictionary
        self.data_dict_second_detail = {
            'securityType': self.securityType,
            'amtSubmitted': self.amtSubmitted,
            'amtAccepted': self.amtAccepted,
            'minimumBidRate': self.minimumBidRate,
            'percentHighRate': self.percentHighRate,
            'percentLowRate': self.percentLowRate,
            'percentStopOutRate': self.percentStopOutRate,
            'percentWeightedAverageRat': self.percentWeightedAverageRat
        }


        self.df = pd.DataFrame(self.data_dict)
        self.detail_df = pd.DataFrame(self.data_dict_second_detail)
    class Details:
        def __init__(self, details):
            self.securityType = [d['securityType'] if 'securityType' in d else None for d in details]
            self.amtSubmitted = [d['amtSubmitted'] if 'amtSubmitted' in d else None for d in details]
            self.amtAccepted = [d['amtAccepted'] if 'amtAccepted' in d else None for d in details]
            self.percentOfferingRate = [d['percentOfferingRate'] if 'percentOfferingRate' in d else None for d in details]
            self.percentAwardRate = [d['percentAwardRate'] if 'percentAwardRate' in d else None for d in details]
            self.minimumBidRate = [d['minimumBidRate'] if 'minimumBidRate' in d else None for d in details]
            self.percentHighRate = [d['percentHighRate'] if 'percentHighRate' in d else None for d in details]
            self.percentLowRate = [d['percentLowRate'] if 'percentLowRate' in d else None for d in details]




class SecuritiesLending:
    def __init__(self, operations):
        self.operationId = [i['operationId'] if 'operationId' in i else None for i in operations]
        self.auctionStatus = [i['auctionStatus'] if 'auctionStatus' in i else None for i in operations]
        self.operationType = [i['operationType'] if 'operationType' in i else None for i in operations]
        self.operationDate = [i['operationDate'] if 'operationDate' in i else None for i in operations]
        self.settlementDate = [i['settlementDate'] if 'settlementDate' in i else None for i in operations]
        self.maturityDate = [i['maturityDate'] if 'maturityDate' in i else None for i in operations]
        self.releaseTime = [i['releaseTime'] if 'releaseTime' in i else None for i in operations]
        self.closeTime = [i['closeTime'] if 'closeTime' in i else None for i in operations]
        self.note = [i['note'] if 'note' in i else None for i in operations]
        self.lastUpdated = [i['lastUpdated'] if 'lastUpdated' in i else None for i in operations]
        self.totalParAmtSubmitted = [i['totalParAmtSubmitted'] if 'totalParAmtSubmitted' in i else None for i in operations]
        self.totalParAmtAccepted = [i['totalParAmtAccepted'] if 'totalParAmtAccepted' in i else None for i in operations]

        self.data_dict = {
            'operationId': self.operationId,
            'auctionStatus': self.auctionStatus,
            'operationType': self.operationType,
            'operationDate': self.operationDate,
            'settlementDate': self.settlementDate,
            'maturityDate': self.maturityDate,
            'releaseTime': self.releaseTime,
            'closeTime': self.closeTime,
            'note': self.note,
            'lastUpdated': self.lastUpdated,
            'totalParAmtSubmitted': self.totalParAmtSubmitted,
            'totalParAmtAccepted': self.totalParAmtAccepted
        }

        self.df = pd.DataFrame(self.data_dict)


import os
from dotenv import load_dotenv
from fudstop.apis.helpers import get_human_readable_string
load_dotenv()

from fudstop.apis.webull.webull_options import WebullOptions
import pandas as pd
options = WebullOptions(connection_string=os.environ.get('WEBULL_OPTIONS'))
import asyncio

async def fetch_and_print_volume_analysis(semaphore, option_id):
    async with semaphore:
        data = await options.fetch_volume_analysis(option_id)
        symbol_query = f"SELECT symbol FROM webull_options WHERE ticker_id = '{option_id}'"
        symbol_result = await options.fetch(symbol_query)
        option_symbol = symbol_result[0]['symbol'] if symbol_result and symbol_result[0].get('symbol') is not None else 'Unknown'
        components = get_human_readable_string(option_symbol)

        strike_price = components.get('strike_price')
        expiry_date = components.get('expiry_date')
        call_put = components.get('call_put')
        symbol = components.get('underlying_symbol')
        data_frame = data.as_dataframe
        data_frame['option_symbol'] = option_symbol
        data_frame['symbol'] = symbol
        data_frame['strike_price'] = strike_price
        data_frame['call_put'] = call_put,
        data_frame['expiry_date'] = expiry_date
        return data_frame

async def main():
    await options.connect()
    option_ids = await options.get_option_id_for_symbol('SPY')
    semaphore = asyncio.Semaphore(35)
    tasks = [fetch_and_print_volume_analysis(semaphore, option_id) for option_id in option_ids]
    dataframes = await asyncio.gather(*tasks)
    final_dataframe = pd.concat(dataframes, ignore_index=True)
    print(final_dataframe)

    final_dataframe.to_csv('SPY_OPTIONS.csv', index=False)





asyncio.run(main())


from dataclasses import dataclass

from typing import Optional, List

import pandas as pd


class TickerNews:
    def __init__(self, data):

        results = data['results']

        self.amp_url = [i['amp_url'] if 'amp_url' is not None and 'amp_url' in i else None for i in results]
        self.article_url = [i['article_url'] if 'article_url' is not None and 'article_url' in i else None for i in results]
        self.author = [i['author'] if 'author' is not None and 'author' in i else None for i in results]
        self.description = [i['description'] if 'description' is not None and 'description' in i else None for i in results]
        self.id = [i['id'] if 'id' is not None and 'id' in i else None for i in results]
        self.image_url = [i['image_url'] if 'image_url' is not None and 'image_url' in i else None for i in results]
        self.keywords = [i['keywords'] if 'keywords' is not None and 'keywords' in i else None for i in results]
        
        publisher = [i['publisher'] if 'publisher' is not None and 'publisher' in i else None for i in results]
        self.tickers = [i['tickers'] if 'tickers' is not None and 'tickers' in i else None for i in results]
        self.favicon_url = [i['favicon_url'] if 'favicon_url' is not None and 'favicon_url' in i else None for i in publisher]
        self.name = [i['name'] if 'name' is not None and 'name' in i else None for i in publisher]
        self.logo_url = [i['favicon_url'] if 'favicon_url' is not None and 'favicon_url' in i else None for i in publisher]
        self.homepage_url = [i['homepage_url'] if 'homepage_url' is not None and 'homepage_url' in i else None for i in publisher]
        self.title = [i['title'] if 'title' is not None and 'title' in i else None for i in results]



        self.data_dict = {
            'Title': self.title,
            'Name': self.name,
            'Author': self.author,
            'Article URL': self.article_url,
            'Homepage URL': self.homepage_url,
            'Logo URL': self.logo_url,
            'Image URL': self.image_url,
            'Tickers Mentioned': self.tickers,
            'Description': self.description,
            'Keywords': self.keywords

        }

        self.as_dataframe = pd.DataFrame(self.data_dict)



from apis.polygonio.async_polygon_sdk import Polygon
import os

import asyncio

poly = Polygon()


# Creating a database configuration dictionary using os.environ.get for environment variables
db_config = {
    "DB_HOST": os.environ.get('DB_HOST', 'localhost'),
    "DB_PASSWORD": os.environ.get('DB_PASSWORD'),
    "DB_NAME": os.environ.get('NEWS_STRING'),
    "DB_PORT": os.environ.get('DB_PORT', 5432),
    "DB_USER": os.environ.get('DB_USER')
}



async def market_news():

    news = await poly.market_news(limit='100')

    print(news.author)
    print(news.as_dataframe)


asyncio.run(market_news())


from fudstop.apis.helpers import convert_to_datetime
from ..weather_helpers import kelvin_to_temp
from fudstop.apis.helpers import flatten_list_of_dicts

import pandas as pd
class WeatherLocation:

    def __init__(self, data, unit:str="F"):
        self.lat = data.get('lat')
        self.lon = data.get('lon')
        self.timezone = data.get('timezone')
        self.timezone_offset =  data.get('timezone_offset')


        current = data.get('current')

        self.current_time = convert_to_datetime(current.get('dt'))
        self.sunrise = convert_to_datetime(current.get('sunrise'))
        self.sunset = convert_to_datetime(current.get('sunset'))
        self.temp = kelvin_to_temp(current.get('temp'),unit=unit)
        self.feels_like = kelvin_to_temp(current.get('feels_like'), unit=unit)
        self.pressure = current.get('pressure')
        self.humidity = current.get('humidity')
        self.dew_point = current.get('dew_point')
        self.uvi = current.get('uvi')
        self.clouds = current.get('clouds')
        self.visibility = current.get('visibility')
        self.wind_speed = current.get('wind_speed')
        self.wind_degree = current.get('wind_deg')
        
        weather = current.get('weather')
        self.weather_summary = [i.get('main') for i in weather]
        self.weather_desc = [i.get('description') for i in weather]

        self.data_dict = {
            'lat': self.lat,
            'lon': self.lon,
            'timezone': self.timezone,
            'timezone_offset': self.timezone_offset,
            'current_time': self.current_time,
            'sunrise': self.sunrise,
            'sunset': self.sunset,
            'temp': self.temp,
            'feels_like': self.feels_like,
            'pressure': self.pressure,
            'humidity': self.humidity,
            'dew_point': self.dew_point,
            'uvi': self.uvi,
            'clouds': self.clouds,
            'visibility': self.visibility,
            'wind_speed': self.wind_speed,
            'wind_degree': self.wind_degree,
            'weather_summary': self.weather_summary,
            'weather_desc': self.weather_desc
        }


        self.as_dataframe = pd.DataFrame(self.data_dict)



import requests
import pandas as pd
session = requests.session()
class WebullVolAnalysis:
    
    def __init__(self, response):

    
        
        # Check if the response JSON contains the required keys
        if all(key in response for key in ('avePrice', 'buyVolume', 'sellVolume', 'nVolume', 'totalVolume')):
            self.avePrice = response.get('avePrice', None)
            self.buyVolume = response.get('buyVolume', None)
            self.sellVolume = response.get('sellVolume', None)
            self.nVolume = response.get('nVolume', None)
            self.totalVolume = response.get('totalVolume', None)
            self.buyPct = (self.buyVolume / self.totalVolume) * 100
            self.sellPct = (self.sellVolume / self.totalVolume) * 100
            self.nPct = (self.nVolume / self.totalVolume) * 100

            self.data_dict = { 

                'avg_traded_price': self.avePrice,
                'buy_volume': self.buyVolume,
                'sell_volume': self.sellVolume,
                'neutral_volume': self.nVolume,
                'total_volume': self.totalVolume,
                'buy_pct': self.buyPct,
                'neutral_pct': self.nPct,
                'sell_pct': self.sellPct
            }

            self.df = pd.DataFrame(self.data_dict, index=[0]).transpose()
        else:
            # If the required keys are not present, set the attribute values to None
            self.avePrice = 0
            self.buyVolume = 0
            self.sellVolume = 0
            self.nVolume = 0
            self.totalVolume = 0
            self.buyPct = 0
            self.sellPct = 0
            self.nPct = 0




"""USE THIS FILE TO GENERATE LISTS OF TICKERS AS NEEDED FOR BATCH CALLS TO FUNCTIONS"""


from apis.webull.webull_markets import WebullMarkets


markets = WebullMarkets()


import asyncio
import pandas as pd



async def generate_tickers_list():
    types = ['totalVolume', 'totalPosition']
    tasks = [markets.get_top_options(type) for type in types]
    results = await asyncio.gather(*tasks)
    all_symbols = []
    
    # Assuming ticker_list is a list of dictionaries
    for entry in results:
        symbols_series = pd.Series(entry.get('symbol'))
        all_symbols.append(symbols_series)
    
    # Concatenate all the Series into one and convert to a list
    all_symbols_concatenated = pd.concat(all_symbols, ignore_index=True)
    symbols_list = all_symbols_concatenated.tolist()
    
    print(symbols_list)
    # Write to a Python file
    with open('list_sets/ticker_lists.py', 'w') as f:
        f.write("most_active_tickers = " + str(list(symbols_list)))

    print(f"Written {len(set(symbols_list))} unique tickers to ticker_lists.py")

    return set(symbols_list)


asyncio.run(generate_tickers_list()) #now you can import this ticker list to run analysis on multiple tickers if needed


import sys
from pathlib import Path
from asyncpg.exceptions import UniqueViolationError
# Add the project directory to the sys.path
project_dir = str(Path(__file__).resolve().parents[2])
if project_dir not in sys.path:
    sys.path.append(project_dir)

from dotenv import load_dotenv
load_dotenv()
from asyncpg import create_pool

import os
from typing import List, Dict

import asyncio
from aiohttp.client_exceptions import ClientConnectorError, ClientOSError, ClientConnectionError, ContentTypeError

from .models.aggregates import AggregatesData
from .models.ticker_news import TickerNews
from .models.company_info import CombinedCompanyResults
from .models.technicals import RSI, EMA, SMA
from .models.ticker_snapshot import StockSnapshot
from datetime import datetime, timedelta
import aiohttp

from urllib.parse import urlencode

from fudstop.apis.helpers import flatten_dict

YOUR_POLYGON_KEY = os.environ.get('YOUR_POLYGON_KEY')

print(YOUR_POLYGON_KEY)

class Polygon:
    def __init__(self, connection_string=None):
        self.connection_string = connection_string
        self.api_key = os.environ.get('YOUR_POLYGON_KEY')
        self.today = datetime.now().strftime('%Y-%m-%d')
        self.yesterday = (datetime.now() - timedelta(days=1)).strftime('%Y-%m-%d')
        self.tomorrow = (datetime.now() + timedelta(days=1)).strftime('%Y-%m-%d')
        self.thirty_days_ago = (datetime.now() - timedelta(days=30)).strftime('%Y-%m-%d')
        self.thirty_days_from_now = (datetime.now() + timedelta(days=30)).strftime('%Y-%m-%d')
        self.fifteen_days_ago = (datetime.now() - timedelta(days=15)).strftime('%Y-%m-%d')
        self.fifteen_days_from_now = (datetime.now() + timedelta(days=15)).strftime('%Y-%m-%d')
        self.eight_days_from_now = (datetime.now() + timedelta(days=8)).strftime('%Y-%m-%d')
        self.eight_days_ago = (datetime.now() - timedelta(days=8)).strftime('%Y-%m-%d')

    async def connect(self, connection_string=None):
            if connection_string:
                self.pool = await create_pool(
                    dsn=connection_string, min_size=1, max_size=10
                )
            else:
                self.pool = await create_pool(
                    host=self.host,
                    port=self.port,
                    user=self.user,
                    password=self.password,
                    database=self.database,
                    min_size=1,
                    max_size=10
                )
            return self.pool

    async def save_structured_message(self, data: dict, table_name: str):
        fields = ', '.join(data.keys())
        values = ', '.join([f"${i+1}" for i in range(len(data))])
        
        query = f'INSERT INTO {table_name} ({fields}) VALUES ({values})'
        print(self.connection_string)
        async with self.pool.acquire() as conn:
            try:
                await conn.execute(query, *data.values())
            except UniqueViolationError:
                print('Duplicate - SKipping')

    async def fetch_endpoint(self, endpoint, params=None):
        # Build the query parameters
        filtered_params = {k: v for k, v in params.items() if v is not None} if params else {}
        
        if filtered_params:
            query_string = urlencode(filtered_params)
            if query_string:
                endpoint = f"{endpoint}?{query_string}&{self.api_key}"
                
        async with aiohttp.ClientSession() as session:
            async with session.get(url=endpoint, params=filtered_params) as response:
                return await response.json()  # or response.text() based on your needs
                



    
    async def aggregates(self, ticker, multiplier:str='1', timespan:str='day', date_from=None, date_to=None, limit:str=500, sort:str='desc'):
        """
        Fetches candlestick data for a ticker, option symbol, crypto/forex pair.
        
        Parameters:
        - ticker (str): The ticker symbol for which to fetch data.

        - timespan: The timespan to survey.

        TIMESPAN OPTIONS:

        >>> second
        >>> minute
        >>> hour
        >>> day
        >>> week
        >>> month
        >>> quarter
        >>> year



        >>> Multiplier: the number of timespans to survey.

        - date_from (str, optional): The starting date for the data fetch in yyyy-mm-dd format.
                                     Defaults to 30 days ago if not provided.
        - date_to (str, optional): The ending date for the data fetch in yyyy-mm-dd format.
                                   Defaults to today's date if not provided.

        - limit: the amount of candles to return. Defaults to 500



        Returns:
        - dict: Candlestick data for the given ticker and date range.

        Example:
        >>> await aggregates('AAPL', date_from='2023-09-01', date_to='2023-10-01')
        """
        
        if date_from is None:
            date_from = (datetime.now() - timedelta(days=30)).strftime('%Y-%m-%d')
        
        if date_to is None:
            date_to = datetime.now().strftime('%Y-%m-%d')
        
  
        params = {
            'adjusted': 'true',
            'sort': sort,
            'limit': limit,
            'apiKey': self.api_key  # API key included here
        }

        endpoint = f"https://api.polygon.io/v2/aggs/ticker/{ticker}/range/{multiplier}/{timespan}/{date_from}/{date_to}"
        
        data = await self.fetch_endpoint(endpoint, params=params)
        data = AggregatesData(data)
        
        return data


    async def market_news(self, limit: str = '100'):
        """
        Arguments:

        >>> ticker: the ticker to query (optional)
        >>> limit: the number of news items to return (optional) | Max 1000

        """
        params = {
            'apiKey': self.api_key,
            'limit': limit
        }


        endpoint = "https://api.polygon.io/v2/reference/news"

        data = await self.fetch_endpoint(endpoint, params=params)
        data = TickerNews(data)

        return data
    

    async def company_info(self, ticker) -> CombinedCompanyResults:
        url = f"https://api.polygon.io/v3/reference/tickers/{ticker}?apiKey={self.api_key}"
        print(url)
        async with aiohttp.ClientSession() as session:
            async with session.get(url) as response:
                data = await response.json()
                results_data = data['results'] if 'results' in data else None
                if results_data is not None:
                    return CombinedCompanyResults(
                        ticker=results_data.get('ticker'),
                        name=results_data.get('name'),
                        market=results_data.get('market'),
                        locale=results_data.get('locale'),
                        primary_exchange=results_data.get('primary_exchange'),
                        type=results_data.get('type'),
                        active=results_data.get('active'),
                        currency_name=results_data.get('currency_name'),
                        cik=results_data.get('cik'),
                        composite_figi=results_data.get('composite_figi'),
                        share_class_figi=results_data.get('share_class_figi'),
                        market_cap=results_data.get('market_cap'),
                        phone_number=results_data.get('phone_number'),
                        description=results_data.get('description'),
                        sic_code=results_data.get('sic_code'),
                        sic_description=results_data.get('sic_description'),
                        ticker_root=results_data.get('ticker_root'),
                        homepage_url=results_data.get('homepage_url'),
                        total_employees=results_data.get('total_employees'),
                        list_date=results_data.get('list_date'),
                        share_class_shares_outstanding=results_data.get('share_class_shares_outstanding'),
                        weighted_shares_outstanding=results_data.get('weighted_shares_outstanding'),
                        round_lot=results_data.get('round_lot'),
                        address1=results_data.get('address', {}).get('address1'),
                        city=results_data.get('address', {}).get('city'),
                        state=results_data.get('address', {}).get('state'),
                        postal_code=results_data.get('address', {}).get('postal_code'),
                        logo_url=results_data.get('branding', {}).get('logo_url'),
                        icon_url=results_data.get('branding', {}).get('icon_url')
                    )
                else:
                    print(f'Couldnt get info for {ticker}')

    async def get_all_tickers(self, include_otc=False, save_all_tickers:bool=False) -> List[StockSnapshot]:
        """
        Fetches a list of all stock tickers available on Polygon.io.

        Arguments:
            >>> include_otc: optional - whether to include OTC securities or not

            >>> save_all_tickers: optional - saves all tickers as a list for later processing

        Returns:
            A list of StockSnapshot objects, each containing data for a single stock ticker.

        Usage:
            To fetch a list of all stock tickers available on Polygon.io, you can call:
            ```
            tickers = await sdk.get_all_tickers()
            print(f"Number of tickers found: {len(tickers)}")
            ```
        """
        url = f"https://api.polygon.io/v2/snapshot/locale/us/markets/stocks/tickers?include_otc={include_otc}&apiKey={self.api_key}"
        params = {
            "apiKey": self.api_key,
        }
        print(url)
        async with aiohttp.ClientSession() as session:
            async with session.get(url, params=params) as response:
                response_data = await response.json()



                tickers = response_data['tickers']
                print(tickers)
                ticker_data = [StockSnapshot(ticker) for ticker in response_data['tickers'] if ticker is not None]

                if save_all_tickers:
                    # Extract tickers to a list
                    ticker_list = [ticker['ticker'] for ticker in tickers]
                    
                    # Write the tickers to a file
                    with open('list_sets/saved_tickers.py', 'w') as f:
                        f.write(str(ticker_list))
                return ticker_data


    async def rsi(self, ticker:str, timespan:str, limit:str='1000', window:int=14, date_from:str=None, date_to:str=None, session=None):
        """
        Arguments:

        >>> ticker

        >>> AVAILABLE TIMESPANS:

        minute
        hour
        day
        week
        month
        quarter
        year

        >>> date_from (optional) 
        >>> date_to (optional)
        >>> window: the RSI window (default 14)
        >>> limit: the number of N timespans to survey
        
        """

        if date_from is None:
            date_from = self.eight_days_ago

        if date_to is None:
            date_to = self.today


        endpoint = f"https://api.polygon.io/v1/indicators/rsi/{ticker}?timespan={timespan}&timestamp.gte={date_from}&timestamp.lte={date_to}&limit={limit}&window={window}&apiKey={self.api_key}"
        print(endpoint)
        async with aiohttp.ClientSession() as session:
            try:
                async with session.get(endpoint) as resp:
                    datas = await resp.json()
                    if datas is not None:

                        
  
                
                        return RSI(datas, ticker)
            except (ClientConnectorError, ClientOSError, ContentTypeError):
                print(f"ERROR - {ticker}")
                


    async def sma(self, ticker:str, timespan:str, limit:str='1000', window:str='9', date_from:str=None, date_to:str=None):
        """
        Arguments:

        >>> ticker

        >>> AVAILABLE TIMESPANS:

        minute
        hour
        day
        week
        month
        quarter
        year

        >>> date_from (optional) 
        >>> date_to (optional)
        >>> window: the SMA window (default 9)
        >>> limit: the number of N timespans to survey
        
        """

        if date_from is None:
            date_from = self.eight_days_ago

        if date_to is None:
            date_to = self.today


        endpoint = f"https://api.polygon.io/v1/indicators/sma/{ticker}?timespan={timespan}&window={window}&timestamp.gte={date_from}&timestamp.lte={date_to}&limit={limit}&apiKey={self.api_key}"

        async with aiohttp.ClientSession() as session:
            async with session.get(endpoint) as resp:
                datas = await resp.json()


                return SMA(datas, ticker)
            


    async def ema(self, ticker:str, timespan:str, limit:str='1', window:str='21', date_from:str=None, date_to:str=None):
        """
        Arguments:

        >>> ticker

        >>> AVAILABLE TIMESPANS:

        minute
        hour
        day
        week
        month
        quarter
        year

        >>> date_from (optional) 
        >>> date_to (optional)
        >>> window: the EMA window (default 21)
        >>> limit: the number of N timespans to survey
        
        """

        if date_from is None:
            date_from = self.eight_days_ago

        if date_to is None:
            date_to = self.today


        endpoint = f"https://api.polygon.io/v1/indicators/ema/{ticker}?timespan={timespan}&window={window}&timestamp.gte={date_from}&timestamp.lte={date_to}&limit={limit}&apiKey={self.api_key}"

        async with aiohttp.ClientSession() as session:
            async with session.get(endpoint) as resp:
                datas = await resp.json()


                return EMA(datas, ticker)
            




    async def get_universal_snapshot(self, ticker, retries=3): #✅
        """Fetches the Polygon.io universal snapshot API endpoint"""
        timeout = aiohttp.ClientTimeout(total=10)  # 10 seconds timeout for the request
        
        for retry in range(retries):
        # async with sema:
            url = f"https://api.polygon.io/v3/snapshot?ticker.any_of={ticker}&apiKey={self.api_key}&limit=250"

            
            async with aiohttp.ClientSession(timeout=timeout) as session:
                try:
                    async with session.get(url) as resp:
                        data = await resp.json()
                        results = data.get('results', None)
        
                        if results is not None:
                            flattened_results = [flatten_dict(result) for result in results]
                            return flattened_results
                            
                except aiohttp.ClientConnectorError:
                    print("ClientConnectorError occurred. Retrying...")
                    continue
                
                except aiohttp.ContentTypeError as e:
                    print(f"ContentTypeError occurred: {e}")  # Consider logging this
                    continue
                
                except Exception as e:
                    print(f"An unexpected error occurred: {e}")  # Consider logging this
                    continue





    async def gather_rsi_for_all_tickers(self, tickers) -> List[RSI]:

        """Get RSI for all tickers
        
        Arguments:

        >>> tickers: A list of tickers


        >>> timespan: 

           minute
           hour
           day
           week
           month
           year
           quaeter
        
        """
        async with aiohttp.ClientSession() as session:
            timespans = ['minute', 'hour', 'day', 'week']
            timespan = 'hour'
            tasks = [self.rsi(ticker, timespan) for ticker in tickers for timespan in timespans]
            await asyncio.gather(*tasks)
                
              
